import logging

import os
from django.http import HttpResponse, HttpResponseBadRequest
from django.shortcuts import render

from golem.nlu.utils import get_entity_names, data_dir
import json


def nlp_view(request):
    context = {'entities': get_entity_names()}
    return render(request, 'nlp/nlp.html', context)


def training_view(request, entity):
    if not entity or not isinstance(entity, str):
        logging.warning("Invalid entity")
        return HttpResponseBadRequest()

    path = os.path.join(data_dir(), 'training_data', entity + '.json')
    with open(path) as f:
        js = json.load(f)
        strategy = js.get("strategy")
        if strategy == "trait":
            return trait_training_view(request, entity, js)
        elif strategy == "keywords":
            return keywords_training_view(request, entity, js)


def trait_training_view(request, entity, js):
    return HttpResponse()


def keywords_training_view(request, entity, js):
    data = js.get("data", [])
    return render(request, "nlp/keywords.html")



import json
import logging
from datetime import datetime, timedelta
from typing import Optional

import requests
from django.conf import settings

from golem.core.interfaces.adapter.microsoft import MicrosoftAdapter
from golem.core.message_parser import parse_text_message
from golem.core.persistence import get_redis
from golem.tasks import accept_user_message


# FIXME needs to be updated

class MicrosoftInterface():
    name = 'microsoft'
    prefix = 'ms'

    @staticmethod
    def get_base_url(chat_id) -> Optional[str]:
        """
        Loads the correct endpoint url for a chat from redis.
        :param chat_id: Id of the chat.
        :return: url, throws Exception if chat_id not valid or url not found
        """
        if not chat_id:
            raise Exception('Chat id must not be null')
        redis = get_redis()
        url = redis.get('ms_service_url_' + str(chat_id))
        if url is None:
            raise Exception('Service url not found for chat id ' + str(chat_id))

        url = url.decode()
        if not url.endswith('/'):
            url += '/'
        url += 'v3/'
        return url

    @staticmethod
    def set_base_url(chat_id, url):
        if not chat_id or not url:
            raise Exception('Chat id and url must not be null')
        redis = get_redis()
        redis.set('ms_service_url_' + str(chat_id), str(url))

    @staticmethod
    def set_bot_id(chat_id, bot_id):
        """
        Because the microsoft bot api doesn't send the requests to recipient "BOT_ID",
        but to recipient "BOT_ID@nnnn" instead and this ID is required to reply,
        it is saved in redis. When MS updates their documentation about IDs this will hopefully be removed.
        :param chat_id:
        :param bot_id:
        :return:
        """
        if chat_id is None or bot_id is None:
            raise Exception('Chat id and bot id must not be null')
        redis = get_redis()
        redis.set('ms_reply_botid_' + str(chat_id), str(bot_id))

    @staticmethod
    def get_bot_id(chat_id):
        if chat_id is None:
            raise Exception('Chat id must not be null')
        redis = get_redis()
        bot_id = redis.get('ms_reply_botid_' + str(chat_id))
        if bot_id is None:
            raise Exception('Bot ID not found for chat {}'.format(str(chat_id)))
        return bot_id.decode()

    @staticmethod
    def clear():
        pass

    @staticmethod
    def load_profile(uid):
        return {'first_name': 'Tests', 'last_name': ''}

    @staticmethod
    def get_auth_token():
        """
        :returns: Auth token for Microsoft Bot API.
        """
        redis = get_redis()
        if redis.exists('ms_token'):
            return redis.get('ms_token').decode()
        else:
            url = 'https://login.microsoftonline.com/botframework.com/oauth2/v2.0/token'
            headers = {
                "Content-Type": "application/x-www-form-urlencoded"
            }
            payload = (
                'grant_type=client_credentials' +
                '&client_id=' + settings.GOLEM_CONFIG.get('MS_BOT_ID') +
                '&client_secret=' + settings.GOLEM_CONFIG.get('MS_BOT_TOKEN') +
                '&scope=' + 'https://api.botframework.com/.default'
            )
            response = requests.post(url, data=payload, headers=headers)
            if response.status_code != 200:
                logging.error(response.text)
                response.raise_for_status()
            auth_data = response.json()
            token = auth_data['access_token']
            ex = auth_data['expires_in']
            redis.set('ms_token', str(token), ex=ex)
            return token

    @staticmethod
    def post_message(uid, chat_id, response):
        if uid and not chat_id:
            # TODO initiate conversation
            return

        message_id = get_redis().get('chat_message_id:{}'.format(chat_id))
        if message_id:
            message_id = message_id.decode()

        bot_id = MicrosoftInterface.get_bot_id(chat_id)

        payload = {
            "type": "message",
            "from": {
                "id": bot_id
            },
            "conversation": {
                "id": chat_id
            },
            "recipient": {
                "id": uid
            },
            "replyToId": message_id
        }
        payload.update(MicrosoftAdapter(chat_id).to_response(response))

        url = MicrosoftInterface.get_base_url(chat_id) + 'conversations/' + chat_id + '/activities'
        if message_id and False:  # replying to message TODO get and clear atomically or set timeout or something
            url += '/' + message_id
        headers = {
            "Authorization": "Bearer " + MicrosoftInterface.get_auth_token(),
            "Content-Type": "application/json"
        }
        logging.warning(url)
        logging.warning(payload)
        response = requests.post(url, data=json.dumps(payload), headers=headers)
        if response.status_code != 200:
            logging.warning(str(payload))
            logging.warning(response)
            logging.warning(response.text)
        response.raise_for_status()

    @staticmethod
    def send_settings(settings):
        pass

    @staticmethod
    def processing_start(uid, chat_id):
        if not chat_id:
            return
        chat_id = str(chat_id)

        payload = {
            "type": "typing",
            "from": {
                "id": MicrosoftInterface.get_bot_id(chat_id)
            },
            "conversation": {
                "id": chat_id
            }
        }
        url = MicrosoftInterface.get_base_url(chat_id) + 'conversations/' + chat_id + '/activities'
        headers = {"Authorization": "Bearer " + MicrosoftInterface.get_auth_token()}
        response = requests.post(url, data=json.dumps(payload), headers=headers)
        if response.status_code != 200:
            logging.error(response.text)
            response.raise_for_status()

    @staticmethod
    def processing_end(uid, chat_id):
        pass

    @staticmethod
    def state_change(state):
        pass

    @staticmethod
    def has_message_expired(message: dict) -> bool:
        if not (message and 'date' in message):
            logging.warning('Invalid date in message')
            return True
        received = datetime.strptime(message['timestamp'], '%Y-%m-%dT%H:%M:%s.%fZ')
        now = datetime.utcnow()

        if abs(now - received) > timedelta(seconds=settings.GOLEM_CONFIG.get('MSG_LIMIT_SECONDS')):
            logging.warning('Ignoring message, too old')
            return True
        return False

    @staticmethod
    def accept_request(body, num_tries=1) -> bool:
        if body['type'] == 'message':
            uid = body['from']['id']
            chat_id = body['conversation']['id']
            get_redis().set('chat_message_id:{}'.format(chat_id), body['id'])  # TODO
            MicrosoftInterface.set_base_url(chat_id, body['serviceUrl'])
            MicrosoftInterface.set_bot_id(chat_id, body['recipient']['id'])
            accept_user_message.delay(MicrosoftInterface.name, uid, body, chat_id=chat_id)
            return True
        return False

    @staticmethod
    def parse_message(raw, num_tries=1):
        if 'text' in raw:
            return parse_text_message(raw['text'])
        elif 'value' in raw:
            payload = raw['value']
            payload['_message_text'] = [{'value': None}]
            return {'entities': payload, 'type': 'postback'}
        return {'type': 'undefined'}



import json
from abc import ABC, abstractmethod

from golem.core.responses.responses import MessageElement
from golem.core.serialize import json_serialize


class Button(MessageElement, ABC):
    """
    An abstract chat button. Don't use this class directly.
    """

    @abstractmethod
    def to_response(self):
        raise NotImplementedError('Abstract method')

    @abstractmethod
    def __str__(self):
        raise NotImplementedError('Abstract method')


class PayloadButton(Button):
    """
    A button that sends predefined data back to the bot when clicked.
    """

    def __init__(self, title, payload):
        super(Button, self).__init__()
        self.title = title
        self.payload = payload

    def to_response(self):
        return {
            'title': self.title,
            'type': 'postback',
            'payload': json.dumps(self.payload, default=json_serialize)
        }

    def __str__(self):
        return 'button: {}: {}'.format(self.title, self.payload)


class LinkButton(Button):
    """
    A button that opens a website when clicked.
    """

    def __init__(self, title, url, webview=False):
        super(Button, self).__init__()
        self.title = title
        self.url = url
        self.webview_height_ratio = 'compact' if webview else 'full'

    def to_response(self):
        return {
            'title': self.title,
            'type': 'web_url',
            'url': self.url,
            'webview_height_ratio': self.webview_height_ratio
        }

    def __str__(self):
        return 'button: {}: {}'.format(self.title, self.url)


class PhoneButton(Button):
    """
    A button that calls a number when clicked.
    Currently only Facebook Messenger is supported.
    """

    def __init__(self, title, phone_number):
        super(Button, self).__init__()
        self.title = title
        self.phone_number = phone_number

    def to_response(self):
        return {
            'title': self.title,
            'type': 'phone_number',
            'payload': self.phone_number
        }

    def __str__(self):
        return 'button: {}: {}'.format(self.title, self.phone_number)


class AccountLinkButton(Button):
    """
    A button that shows the FB messenger account linking dialog when clicked.
    """

    def __init__(self, url):
        super(Button, self).__init__()
        self.url = url

    def to_response(self):
        return {
            'type': 'account_link',
            'url': self.url
        }

    def __str__(self):
        return 'button: account_link: {}'.format(self.url)


class AccountUnlinkButton(Button):
    """
    A button that shows the FB messenger account unlinking dialog when clicked.
    """

    def __init__(self, url):
        super(Button, self).__init__()
        self.url = url

    def to_response(self):
        return {
            'type': 'account_unlink'
        }

    def __str__(self):
        return 'button: account_unlink'


class ShareButton(Button):
    """
    A button that opens share dialog for the message it's attached to.
    Currently only Facebook Messener is supported.
    """

    def __init__(self):
        super(Button, self).__init__()

    def to_response(self):
        return {
            'type': 'element_share'
        }

    def __str__(self):
        return 'button: share'



import re

from .templates import Templates


class Flow:
    def __init__(self, name, dialog, definition):
        self.name = name
        self.dialog = dialog
        self.states = {}
        self.current_state_name = 'root'
        self.intent = definition.get('intent') or name
        for state_name,state_definition in definition['states'].items():
            self.states[state_name] = State(name+'.'+state_name, dialog=dialog, definition=state_definition)

    def get_state(self, state_name):
        return self.states.get(state_name)

    def __str__(self):
        return self.name + ":flow"

class State:
    def __init__(self, name, dialog, definition):
        from .dialog_manager import DialogManager
        self.name = name  # type: str
        self.dialog: DialogManager = dialog  # type: DialogManager
        self.intent_transitions = definition.get('intent_transitions') or {}
        self.intent = definition.get('intent')
        self.init = self.create_action(definition.get('init'))
        self.accept = self.create_action(definition.get('accept'))

    def create_action(self, definition):
        if not definition:
            return None
        if callable(definition):
            return definition
        template = definition.get('template')
        params = definition.get('params') or None
        if hasattr(Templates, template):
            fn = getattr(Templates, template)
        else:
            raise ValueError('Template %s not found, create a static method Templates.%s' % (template))

        return fn(**params)

    def get_intent_transition(self, intent):
        for key,state_name in self.intent_transitions.items():
            if re.match(key, intent): return state_name
        return None

    def __str__(self):
        return self.name + ":state"

    def __repr__(self):
        return str(self)

def require_one_of(entities=[]):
    def decorator_wrapper(func):
        def func_wrapper(state):
            all_entities = entities + ['intent','_state']
            if not state.dialog.context.has_any(all_entities, max_age=0):
                print('No required entities present, moving to default.root: {}'.format(all_entities))
                return None, 'default.root:accept'
            return func(state)
        return func_wrapper
    return decorator_wrapper


import json
import logging

import random

from golem.core.chat_session import ChatSession
from golem.core.message_parser import parse_text_message
from golem.tasks import accept_user_message
from .models import WebMessageData


class WebGuiInterface:
    name = 'webgui'
    prefix = 'web'
    messages = []
    states = []

    @staticmethod
    def clear():
        WebGuiInterface.messages = []
        WebGuiInterface.states = []

    @staticmethod
    def load_profile(uid):
        return {'first_name': 'Tests', 'last_name': ''}

    @staticmethod
    def post_message(session, response):
        uid = session.meta.get("uid")
        WebGuiInterface.messages.append(response)
        message = WebMessageData()
        message.uid = uid
        message.is_response = True
        message.data = json.dumps(response.to_response())
        message.save()

    @staticmethod
    def send_settings(settings):
        pass

    @staticmethod
    def processing_start(session):
        pass

    @staticmethod
    def processing_end(session):
        pass

    @staticmethod
    def state_change(state):
        if not WebGuiInterface.states or WebGuiInterface.states[-1] != state:
            WebGuiInterface.states.append(state)

    @staticmethod
    def parse_message(user_message, num_tries=1):
        logging.info('[WEBGUI] @ parse_message')
        if user_message.get('text'):
            return parse_text_message(user_message.get('text'))
        elif user_message.get("payload"):
            # data = json.loads(user_message["payload"])
            data = user_message['payload']
            # data = json.loads(b64decode(user_message["payload"]).decode())
            logging.info("Payload is: {}".format(data))
            if isinstance(data, dict):
                return {'entities': data, 'type': 'postback'}
            else:
                from golem.core.serialize import json_deserialize
                payload = json.loads(data, object_hook=json_deserialize)
                payload['_message_text'] = [{'value': None}]
                return {'entities': payload, 'type': 'postback'}

    @staticmethod
    def accept_request(msg: WebMessageData):
        uid = str(msg.uid)

        logging.info('[WEBGUI] Received message from {}'.format(uid))
        session = ChatSession(WebGuiInterface, uid, meta={"uid": uid})
        accept_user_message.delay(session.to_json(), {"text": msg.message().get('text')})

    @staticmethod
    def accept_postback(msg: WebMessageData, data):
        uid = str(msg.uid)
        data = json.loads(data)

        logging.info('[WEBGUI] Received postback from {}'.format(uid))
        session = ChatSession(WebGuiInterface, uid, meta={"uid": uid})
        accept_user_message.delay(session.to_json(), {"_message_text": msg.message().get('text'), "payload": data})

    @staticmethod
    def make_uid(username) -> str:
        uid = None
        tries = 0
        while (not uid or len(WebMessageData.objects.filter(uid__exact=uid)) != 0) or tries < 100:
            uid = '_'.join([str(username), str(random.randint(1000, 99999))])
            tries += 1
        # delete messages for old session with this uid, if there was one
        # FIXME invalidate the previous user's session!
        try:
            WebMessageData.objects.get(uid__exact=uid).delete()
        except Exception:
            pass  # first user, there is no such table yet
        return uid

    @staticmethod
    def destroy_uid(uid):
        WebMessageData.objects.filter(uid__exact=uid).delete()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Button',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('text', models.CharField(max_length=255, blank=True, null=True)),
                ('action', models.CharField(max_length=255, blank=True, null=True)),
                ('url', models.CharField(max_length=1024, blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Element',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('title', models.CharField(max_length=255, blank=True, null=True)),
                ('subtitle', models.CharField(max_length=255, blank=True, null=True)),
                ('image_url', models.CharField(max_length=255, blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('text', models.CharField(max_length=255)),
                ('uid', models.CharField(max_length=255)),
                ('timestamp', models.BigIntegerField()),
                ('is_response', models.BooleanField(default=False)),
            ],
        ),
        migrations.AddField(
            model_name='element',
            name='message',
            field=models.ForeignKey(to='webgui.Message', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='button',
            name='message',
            field=models.ForeignKey(to='webgui.Message', on_delete=models.CASCADE),
        ),
    ]



from golem.core.responses.responses import MenuElement


class ThreadSetting():
    def to_response(self):
        pass


class GreetingSetting(ThreadSetting):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return 'greeting_setting'


class GetStartedSetting(ThreadSetting):
    def __init__(self, payload):
        self.payload = payload

    def __str__(self):
        return 'get_started_setting'


class MenuSetting(ThreadSetting):
    def __init__(self, elements=None):
        self.elements = [MenuElement(**element) for element in elements or []]

    def __str__(self):
        text = 'menu:'
        for element in self.elements:
            text += "\n " + str(element)
        return text

    def add_element(self, element):
        self.elements.append(element)
        return element

    def create_element(self, **kwargs):
        element = MenuElement(**kwargs)
        return self.add_element(element)



import emoji
import re
from django.conf import settings

ENTITY_EXTRACTORS = settings.GOLEM_CONFIG.get("ENTITY_EXTRACTORS", [])


def add_default_extractors():
    # compatibility for old chatbots
    if "WIT_TOKEN" in settings.GOLEM_CONFIG:
        print("Adding wit extractor from wit token")
        from golem.core.parsing import wit_extractor
        ENTITY_EXTRACTORS.append(wit_extractor.WitExtractor(settings.GOLEM_CONFIG['WIT_TOKEN']))


add_default_extractors()


def parse_text_message(text, num_tries=1):
    if len(ENTITY_EXTRACTORS) <= 0:
        print('No entity extractors configured!')
        return {'type': 'message', 'entities': {'_message_text': {'value': text}}}

    entities = {}

    for extractor in ENTITY_EXTRACTORS:
        append = extractor.extract_entities(text)
        for entity, values in append.items():
            entities.setdefault(entity, []).extend(values)

    print('Extracted entities:', entities)
    append = parse_additional_entities(text)

    for (entity, value) in re.findall(re.compile(r'/([^/]+)/([^/]+)/'), text):
        if not entity in append:
            append[entity] = []
        append[entity].append({'value': value})

    # add all new entity values
    for entity, values in append.items():
        if not entity in entities:
            entities[entity] = []
        entities[entity] += values

    entities['_message_text'] = [{'value': text}]

    parsed = {'entities': entities, 'type': 'message'}

    return parsed


def parse_additional_entities(text):
    # TODO custom nlp might receive them preprocessed (or not)
    entities = {}
    chars = {':)': ':slightly_smiling_face:', '(y)': ':thumbs_up_sign:', ':(': ':disappointed_face:',
             ':*': ':kissing_face:', ':O': ':face_with_open_mouth:', ':D': ':grinning_face:',
             '<3': ':heavy_black_heart:️', ':P': ':face_with_stuck-out_tongue:'}
    demojized = emoji.demojize(text)
    char_emojis = re.compile(
        r'(' + '|'.join(chars.keys()).replace('(', '\(').replace(')', '\)').replace('*', '\*') + r')')
    demojized = char_emojis.sub(lambda x: chars[x.group()], demojized)
    if demojized != text:
        match = re.compile(r':([a-zA-Z_0-9]+):')
        for emoji_name in re.findall(match, demojized):
            if 'emoji' not in entities:
                entities['emoji'] = []
            entities['emoji'].append({'value': emoji_name})
        # if re.match(match, demojized):
        #    entities['intent'].append({'value':'emoji'})
    return entities



import os


def install_skeleton(dest_dir):
    if not os.path.isdir(dest_dir):
        raise ValueError("{} is not a directory".format(dest_dir))

    src_dir = os.path.dirname(__file__)
    app_name = os.path.basename(os.path.normpath(dest_dir))

    for root, dirs, files in os.walk(src_dir):

        rel_dir = os.path.relpath(root, src_dir)

        for name in dirs:
            dest = os.path.join(dest_dir, rel_dir, name)
            print("Create", dest)
            os.mkdir(dest, 0o755)

        for name in files:
            if not name.endswith('-tpl'):
                continue
            new_name = name.rsplit('-tpl', 1)[0]

            src = os.path.join(src_dir, rel_dir, name)
            dest = os.path.join(dest_dir, rel_dir, new_name)
            print("Copy", src, dest)

            with open(src, 'r') as f:
                file_content = f.read().replace("{APP_NAME}", app_name)
            with open(dest, 'w') as f:
                f.write(file_content)

    print("Adding settings.py imports ...")
    with open(os.path.join(dest_dir, "settings.py"), "r") as f:
        settings_content = f.read()
    # \n is converted by stdlib, keep calm
    settings_content = "from .golm_settings import *\n" + settings_content
    apps_str = "\n    'golem',\n    'webgui',"  # FIXME golem-webgui
    idx = settings_content.find("INSTALLED_APPS")
    idx = settings_content.find("\n", idx)
    settings_content = settings_content[:idx] + apps_str + settings_content[idx:]
    with open(os.path.join(dest_dir, "settings.py"), "w") as f:
        f.write(settings_content)



import os
import pandas as pd
import pickle
import requests
import tarfile
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import accuracy_score
from sklearn.model_selection import train_test_split
from sklearn.multiclass import OneVsRestClassifier

from golem.core.parsing.entity_extractor import EntityExtractor


class LanguageDetector(EntityExtractor):
    def __init__(self):
        super().__init__()
        try:
            with open("data/nlp/lang/lang_det.pkl", 'rb') as f:
                self.tfidf, self.cls = pickle.load(f)
        except Exception:
            self.tfidf, self.cls = self.train()

    def load_training_data(self):
        dl_path = os.path.join('data/nlp/lang', "sentences.tar.bz2")
        ds_path = os.path.join('data/nlp/lang', 'sentences.csv')

        if not os.path.isfile(ds_path):
            if not os.path.exists('data/nlp/lang'):
                os.mkdir('data/nlp/lang')
            print("Downloading language dataset ...")
            response = requests.get('http://downloads.tatoeba.org/exports/sentences.tar.bz2', stream=True)
            response.raise_for_status()
            with open(dl_path, 'wb') as f:
                for buf in response.iter_content(4096):
                    f.write(buf)

            print("Unzipping language dataset ...")
            with tarfile.open(dl_path, 'r:bz2') as f:
                f.extractall(path='data/nlp/lang')

        df = pd.read_csv(ds_path, delimiter='\t', header=None, error_bad_lines=False, warn_bad_lines=False)
        df.columns = ['id', 'lang', 'text']
        df = df[df['lang'].isin(['ces', 'slk', 'eng'])]  # TODO
        return df

    def train(self):
        print("Training language detector")
        df = self.load_training_data()
        # df.head()
        languages = df['lang'].drop_duplicates().values
        print(languages)
        lang_to_id = dict((lang, idx) for idx, lang in enumerate(languages))
        x_train, x_test, y_train, y_test = train_test_split(df['text'], df['lang'], test_size=0.1)
        tfidf = TfidfVectorizer().fit(x_train)
        # train
        cls = OneVsRestClassifier(LogisticRegression())
        cls.fit(tfidf.transform(x_train), y_train)  # TODO unidecode?
        y_pred = cls.predict(tfidf.transform(x_train))
        print("Training accuracy:", accuracy_score(y_train, y_pred))
        # test
        y_pred = cls.predict(tfidf.transform(x_test))
        print("Testing accuracy:", accuracy_score(y_test, y_pred))
        print("Language detector trained")
        with open("data/nlp/lang/lang_det.pkl", 'wb') as f:
            pickle.dump([tfidf, cls], f)
        return tfidf, cls

    def extract_entities(self, text: str, max_retries=5):
        features = self.tfidf.transform([text])
        labels = self.cls.predict(features)
        return {"_language": [{"value": labels[0]}]}



from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Trains the NLP for text entity extraction'

    def add_arguments(self, parser):
        parser.add_argument('entity', nargs='+', type=str)

    def handle(self, *args, **options):
        from golem.nlp import train
        if options.get('entity'):
            if options['entity'][0] == 'all':
                train.train_all()
            train.train_all(options['entity'])
        else:
            raise Exception()



import json
import time
import random

from golem.core.persistence import get_redis
from golem.core.responses.responses import *
from golem.core.responses.buttons import *
from django.conf import settings
from .interfaces.test import TestInterface
from .message_parser import parse_text_message
from .serialize import json_deserialize
from .chat_session import ChatSession


class ConversationTestException(Exception):
    def __init__(self, message):
        super(ConversationTestException, self).__init__(message)

class UserMessage:
    def get_parsed(self):
        pass

class UserTextMessage(UserMessage):
    def __init__(self, text):
        self.text = text
        self.entities = {}

    def produces_entity(self, entity, value = None):
        self.entities[entity] = value
        return self

    def get_parsed(self):
        TestLog.log('Sending user message "{}".'.format(self.text))
        parsed = parse_text_message(self.text)

        for entity, values in parsed['entities'].items():
            TestLog.log('- Entity {}: {}'.format(entity, values))

        for entity in self.entities:
            expected_value = self.entities[entity]
            if entity not in parsed['entities']:
                raise ConversationTestException('- Expected entity "{}" not extracted from message "{}".'.format(entity, self.text))
            elif expected_value:
                value = parsed['entities'][entity][0]['value']
                if expected_value != value:
                    raise ConversationTestException('- Extracted "{}" instead of "{}" in entity "{}" from message "{}".'.format(value, expected_value, entity, self.text))
                TestLog.log('- Extracted expected value "{}" of entity "{}".'.format(value, entity))
            else:
                TestLog.log('- Extracted expected entity "{}".'.format(entity))

        return parsed

class UserPostbackMessage(UserMessage):
    def __init__(self, payload):
        self.payload = payload

    def get_parsed(self):
        TestLog.log('Sending user postback: {}.'.format(self.payload))
        return {'entities': self.payload, 'type':'postback'}


class UserButtonMessage(UserMessage):
    def __init__(self, title):
        self.title = title


class BotMessage():
    def __init__(self, klass):
        self.klass = klass
        self.text = None
        self.can_be_repeated = False

    def with_text(self, text):
        self.text = text
        return self

    def repeated(self):
        self.can_be_repeated = True
        return self

    def check(self, message):
        TestLog.log('Checking message: {}'.format(message))
        
        if message is None:
            raise ConversationTestException('- Expected message type {} but no message returned by bot.'.format(self.klass.__name__))
        if not isinstance(message, self.klass):
            raise ConversationTestException('- Expected message type {} but received {} from bot: "{}".'.format(self.klass.__name__, type(message).__name__, message))
        
        TestLog.log('- Checking received {} from bot: "{}".'.format(self.klass.__name__, message))

        if self.text:
            if not isinstance(message, TextMessage):
                raise ConversationTestException('- Only text messages can be checked for text content, but received {}.'.format(type(message)))
            if self.text != message.text:
                raise ConversationTestException('- Expected text content "{}" but received "{}"'.format(self.text, message.text))
    
            TestLog.log('- Received expected message text "{}" from bot.'.format(self.text))

        return True

class StateChange():
    def __init__(self, name):
        self.name = name

    def check(self, state):
        TestLog.log('Checking state: {}'.format(state))
        if state is None:
            raise ConversationTestException('Expected new state {} but no state change occurred.'.format(self.name))
        if self.name != state:
            raise ConversationTestException('Expected new state {} but moved to {}.'.format(self.name, state))

        TestLog.log('Moved to expected new state {}.'.format(self.name))
        return True



class ConversationTest:
    def __init__(self, name, actions, benchmark=False):
        self.benchmark = benchmark
        self.chat_id = 'benchmark_'+str(random.randint(1, 10000000)) if benchmark else 'test'
        self.name = name
        self.actions = actions
        self.session = ChatSession(unique_id=self.chat_id, interface=TestInterface, is_logged=not self.benchmark, is_test=True)

    def run(self):
        self.init()
        stats = []
        for action in self.actions:
            stat = self.run_action(action)
            if stat:
                stats.append(stat)
        
        report = {'min':{'total':0}, 'max':{'total':0}, 'avg':{'total':0}}
        for key in ['init','parsing','processing']:
            report['min'][key] = min([s[key] for s in stats])
            report['avg'][key] = sum([s[key] for s in stats]) / len(stats)
            report['max'][key] = max([s[key] for s in stats])
            report['min']['total'] += report['min'][key]
            report['avg']['total'] += report['avg'][key]
            report['max']['total'] += report['max'][key]
        return report

    def init(self):
        self.buttons = {}
        from .dialog_manager import DialogManager
        DialogManager.clear_chat(self.session.chat_id)
        TestLog.clear()
        TestInterface.clear()
        

    def run_action(self, action):
        
        if isinstance(action, UserButtonMessage):
            payload = self.buttons.get(action.title, {}).get('payload')
            if not payload:
                raise ConversationTestException('Wanted to press button {}, but it is not present.'.format(action.title))
            action = UserPostbackMessage(payload)

        if isinstance(action, UserMessage):
            from .dialog_manager import DialogManager
            start_time = time.time() # test_id=self.name, use_logging=not self.benchmark
            dialog = DialogManager(session=self.session)
            time_init = time.time() - start_time
            start_time = time.time()
            parsed = action.get_parsed()
            time_parsing = time.time() - start_time
            start_time = time.time()
            dialog.process(parsed['type'], parsed['entities'])
            time_processing = time.time() - start_time
            return {'init':time_init, 'parsing':time_parsing, 'processing':time_processing}

        if isinstance(action, BotMessage):
            message = TestInterface.messages.pop(0) if TestInterface.messages else None
            action.check(message)
            if action.can_be_repeated:
                while TestInterface.messages:
                    peek = TestInterface.messages[0]
                    try:
                        action.check(peek)
                        print('Message matches repeatable message, popping: {}'.format(peek))
                        message = TestInterface.messages.pop(0)
                    except:
                        break
            buttons = []
            if isinstance(message, TextMessage):
                buttons = message.buttons
                buttons += message.quick_replies
            if isinstance(message, GenericTemplateMessage):
                for element in message.elements:
                    buttons += element.buttons
            for button_qr in buttons:
                if hasattr(button_qr, 'payload') and button_qr.payload:
                    button_qr.payload['_log_text'] = button_qr.title
                print('Adding button {}'.format(button_qr.title))
                self.buttons[button_qr.title] = {'payload':getattr(button_qr, 'payload', None), 'url': getattr(button_qr, 'url', None)}

        if isinstance(action, StateChange):
            state = TestInterface.states.pop(0) if TestInterface.states else None
            action.check(state)

        return None



class ConversationTestRecorder:

    @staticmethod
    def record_user_message(message_type, message):
        print('Recording user {}: {}'.format(message_type, message))
        db = get_redis()
        db.lpush('test_actions', json.dumps({'type':'user', 'body':{'type':message_type,'entities':message}}, default=json_serialize))

    @staticmethod
    def record_bot_message(message):
        record = {'type':type(message).__name__}
        if isinstance(message, TextMessage):
            record['text'] = message.text
        print('Recording bot message: {}'.format(record))
        db = get_redis()
        db.lpush('test_actions', json.dumps({'type':'bot', 'body':record}))

    @staticmethod
    def record_state_change(state):
        print('Recording state change: {}'.format(state))
        db = get_redis()
        db.lpush('test_actions', json.dumps({'type':'state', 'body':state}))

    @staticmethod
    def record_start():
        print('Starting recording')
        db = get_redis()
        db.delete('test_actions')
        return TextMessage("Starting recording ;)", buttons=[{'title':'Stop recording', 'payload':{'test_record':'stop'}}])

    @staticmethod
    def record_stop():
        print('Stopping recording')
        responses = []
        responses.append(TextMessage("Done recording ;)", buttons=[{'title':'Get result', 'url':settings.GOLEM_CONFIG.get('DEPLOY_URL')+'/golem/test_record'}, {'title':'Start again', 'payload':{'test_record':'start'}}]))
        return responses

    @staticmethod
    def get_result():
        db = get_redis()
        actions = db.lrange('test_actions', 0, -1)
        response = """from golem.core.responses import *
from golem.core.tests import *

actions = []
"""
        state = None
        for action in actions[::-1]:
            action = json.loads(action.decode('utf-8'), object_hook=json_deserialize)
            body = action['body']
            print('Action body: {}'.format(body))
            if action['type'] == 'user':
                if body['type'] == 'message':
                    text_entity = body['entities']['_message_text'][0]
                    text = text_entity['value']
                    message = 'UserTextMessage("{}")'.format(text)
                    for entity in body['entities']:
                        if entity.startswith('_'):
                            continue
                        value = body['entities'][entity][0].get('value')
                        if isinstance(value, str):
                            message += '.produces_entity("{}","{}")'.format(entity, value)
                        else:
                            message += '.produces_entity("{}")'.format(entity)
                if body['type'] == 'postback':
                    text_entity = body['entities'].get('_message_text')[0]
                    text = text_entity['value']
                    message = 'UserButtonMessage("{}")'.format(text)
                response += "\n\n"+'actions.append({})'.format(message) + "\n"
            elif action['type'] == 'bot':
                message = 'BotMessage({})'.format(body['type'])
                if 'text' in body:
                    message += '.with_text("{}")'.format(body['text'])
                response += 'actions.append({})'.format(message)
            elif action['type'] == 'state':
                if body == state:
                    continue
                response += 'actions.append(StateChange("{}"))'.format(body)
                state = body
            response += "\n"

        return response
        
        
class TestLog:
    messages = []

    @staticmethod
    def clear():
        TestLog.messages = []

    @staticmethod
    def log(message):
        print(message)
        TestLog.messages.append(message)

    @staticmethod
    def get():
        return TestLog.messages


import datetime
from datetime import timedelta

import dateutil.parser
import pytz
from celery.utils.log import get_task_logger
from dateutil.relativedelta import relativedelta
from django.utils import timezone

logger = get_task_logger(__name__)


# TODO i18n


def process_datetime(values, duration=None):
    """
    Output example:
        Q: next_week: (datetime.datetime(2016, 11, 21, 0, 0, tzinfo=tzoffset(None, 3600)), datetime.datetime(2016, 11, 28, 0, 0, tzinfo=tzoffset(None, 3600)))
        Q: tomorrow: (datetime.datetime(2016, 11, 18, 0, 0, tzinfo=tzoffset(None, 3600)), datetime.datetime(2016, 11, 19, 0, 0, tzinfo=tzoffset(None, 3600)))
        Q: tonight: (datetime.datetime(2016, 11, 17, 18, 0, tzinfo=tzoffset(None, 3600)), datetime.datetime(2016, 11, 18, 0, 0, tzinfo=tzoffset(None, 3600)))
        Q: at weekend: (datetime.datetime(2016, 11, 18, 18, 0, tzinfo=tzoffset(None, 3600)), datetime.datetime(2016, 11, 21, 0, 0, tzinfo=tzoffset(None, 3600)))
    """

    append = {'date_interval': []}
    for value in values:
        try:
            formatted = None
            if value['type'] == 'interval':
                if 'from' not in value:  # default interval start is now
                    date_from = datetime.datetime.now().replace(tzinfo=pytz.timezone('Europe/Prague'))
                else:
                    date_from = dateutil.parser.parse(value['from']['value'])
                date_to = dateutil.parser.parse(value['to']['value']) - timedelta(seconds=1)
                grain = value['from']['grain']
            else:
                grain = value['grain']
                date_from = dateutil.parser.parse(value['value'])
                if grain == 'week' and date_from == date_this_week(date_from.tzinfo):
                    # change "this week" to next 7 days
                    date_from = timezone.now()
                    formatted = 'the next seven days'
                date_to = date_from + timedelta_from_grain(grain)
                if 'datetime' not in append:
                    append['datetime'] = []
                append['datetime'].append({'value': date_from, 'grain': grain})
            if not formatted:
                formatted = format_date_interval(date_from, date_to, grain)
            append['date_interval'].append({'value': (date_from, date_to), 'grain': grain, 'formatted': formatted})
        except ValueError:
            logger.exception('Error parsing date: {}'.format(value))
    return append


def timedelta_from_grain(grain):
    if grain == 'second':
        return timedelta(seconds=1)
    if grain == 'minute':
        return timedelta(minutes=1)
    if grain == 'hour':
        return timedelta(hours=1)
    if grain == 'day':
        return timedelta(days=1)
    if grain == 'week':
        return timedelta(days=7)
    if grain == 'month':
        return timedelta(days=31)
    if grain == 'year':
        return timedelta(days=365)
    return timedelta(days=1)


def date_now(tzinfo):
    return datetime.datetime.now(tzinfo)


def date_today(tzinfo):
    return date_now(tzinfo).replace(hour=0, minute=0, second=0, microsecond=0)


def date_this_week(tzinfo):
    today = date_today(tzinfo)
    return today - timedelta(days=today.weekday())


def date_this_month(tzinfo):
    today = date_today(tzinfo)
    return today - timedelta(days=today.day - 1)


def format_date_interval(from_date, to_date, grain):
    tzinfo = from_date.tzinfo
    now = date_now(tzinfo)
    today = date_today(tzinfo)
    this_week = date_this_week(tzinfo)
    next_week = this_week + timedelta(days=7)
    this_month = date_this_month(tzinfo)
    next_month = date_this_month(tzinfo) + relativedelta(months=1)

    diff_hours = (to_date - from_date).total_seconds() / 3600
    logger.info('Diff hours: %s' % diff_hours)

    if grain in ['second', 'minute'] and (now - from_date).total_seconds() < 60 * 5:
        return 'now'

    for i in range(0, 6):
        # if the dates are within the i-th day
        if from_date >= today + timedelta(days=i) and to_date <= today + timedelta(days=i + 1):
            if i == 0:
                day = 'today'
            elif i == 1:
                day = 'tomorrow'
            else:
                day = '%s' % from_date.strftime("%A")
            if from_date.hour >= 17:
                return 'this evening' if i == 0 else day + ' evening'
            if from_date.hour >= 12:
                return 'this afternoon' if i == 0 else day + ' afternoon'
            if to_date.hour >= 0 and to_date.hour < 13 and to_date.hour > 0:
                return 'this morning' if i == 0 else day + ' morning'
            return day

    if from_date == this_week and to_date == next_week:
        return 'this week'

    if from_date == next_week and to_date == next_week + timedelta(days=7):
        return 'next week'

    if from_date == this_month and to_date == next_month:
        return 'this month'

    if diff_hours <= 25:  # (25 to incorporate possible time change)
        digit = from_date.day % 10
        date = 'the {}{}'.format(from_date.day, 'st' if digit == 1 else ('nd' if digit == 2 else 'th'))
        return date if from_date.month == now.month else date + ' ' + from_date.strftime('%B')
    return 'these dates'



import logging

interfaces = []


def register_chat_interface(interface):
    """
    Registers a chat interface.
    :param interface: Class of the interface to register.
                      See golem.core.interfaces for reference.
    """
    interfaces.append(interface)


def get_interfaces():
    """
    :returns: List of all registered chat interface classes.
    """
    from golem.core.interfaces.facebook import FacebookInterface
    from golem.core.interfaces.telegram import TelegramInterface
    from golem.core.interfaces.microsoft import MicrosoftInterface
    from golem.core.interfaces.google import GoogleActionsInterface
    from golem.core.interfaces.test import TestInterface
    return interfaces + [FacebookInterface, TelegramInterface, MicrosoftInterface, GoogleActionsInterface,
                         TestInterface]


def create_from_name(name):
    ifs = get_interfaces()
    for interface in ifs:
        if interface.name == name:
            return interface


def create_from_prefix(prefix):
    ifs = get_interfaces()
    for interface in ifs:
        if interface.prefix == prefix:
            return interface


def create_from_chat_id(chat_id):
    prefix = chat_id.split("_", maxsplit=1)[0]
    return create_from_prefix(prefix)


def init_webhooks():
    """
    Registers webhooks for telegram messages.
    """
    logging.debug('Trying to register telegram webhook')
    try:
        from golem.core.interfaces.telegram import TelegramInterface
        TelegramInterface.init_webhooks()
    except Exception as e:
        logging.exception('Couldn\'t init webhooks')


def uid_to_interface_name(uid: str):
    prefix = str(uid).split('_')[0]

    for i in get_interfaces():
        if i.prefix == prefix:
            return i.name
    raise Exception('No interface for {}'.format(uid))



from golem.core.interfaces.all import create_from_prefix


class ChatSession:

    def __init__(self, interface, unique_id: str, meta=None, profile=None, is_logged=True, is_admin=False, is_test=False):
        """
        :param interface: Chat interface object
        :param unique_id: A string that uniquely identifies the user within the interface
        :param profile: Profile object
        :param meta: Any interface-specific data. type: dict
        """
        if interface is None:
            raise ValueError("Interface must be set")
        elif not isinstance(unique_id, str):
            raise ValueError("UID must be a string")
        elif not isinstance(profile, Profile) and profile is not None:
            raise ValueError("Profile must be instance of class Profile")
        self.interface = interface
        self.chat_id = self.interface.prefix + "_" + unique_id
        self.profile = profile or Profile()
        self.meta = meta or {}
        self.is_logged = is_logged
        self.is_admin = is_admin
        self.is_test = is_test

    def to_json(self):
        return {
            "interface": self.interface.prefix,
            "unique_id": self.chat_id.split('_', maxsplit=1)[1],
            "profile": self.profile.to_json(),
            "meta": self.meta,
            "is_logged": self.is_logged,
            "is_admin": self.is_admin
        }

    @staticmethod
    def from_json(data: dict):
        interface = create_from_prefix(data['interface'])
        return ChatSession(
            interface=interface,
            unique_id=data['unique_id'],
            profile=Profile.from_json(data['profile']),
            meta=data['meta'],
            is_logged=data['is_logged'],
            is_admin=data['is_admin']
        )


class Profile:
    def __init__(self, profile_id=None, first_name=None, last_name=None):
        self.profile_id = profile_id
        self.first_name = first_name
        self.last_name = last_name

    def to_json(self):
        return {
            "profile_id": self.profile_id,
            "first_name": self.first_name,
            "last_name": self.last_name
        }

    @staticmethod
    def from_json(data: dict):
        return Profile(
            profile_id=data['profile_id'],
            first_name=data['first_name'],
            last_name=data['last_name']
        )



from django.apps import AppConfig


class GolemConfig(AppConfig):
    name = 'golem'

    def ready(self):
        print('Init webhooks @ GolemConfig')
        from golem.core.interfaces.all import init_webhooks
        init_webhooks()



import re

from golem.nlu.tools.czech_stemmer import cz_stem_all
from golem.nlu.tools.porter2 import Stemmer
from golem.nlu.tools.toktok import ToktokTokenizer

tokenizer = ToktokTokenizer()

stemmers = {
    "en": Stemmer("en").stemWords,
    "cz": cz_stem_all
}


def tokenize(text: str, stemming=False, language='en'):
    """
    Tokenizes text, optionally stemming the tokens.
    :param text:        Text to be tokenized.
    :param stemming:    Boolean, enables stemming.
                        Default is false.
    :param language:    Stemming language. Default is 'en'.
    :return:
    """
    tokens = tokenizer.tokenize(text)
    if stemming:
        stemmer = stemmers.get(language)
        if not stemmer:
            raise Exception("No known stemmer for language {}!".format(language))
        tokens = [s.lower() for s in stemmer(tokens)]
    return [str(token).lower() for token in tokens]


def build_imputation_rules(imputation: dict):
    rules = []
    for item in imputation:
        stem, suffixes, nearest = item['stem'], item['suffixes'], item['nearest']
        regex = re.escape(stem) + "(?:" + "|".join([re.escape(suffix) for suffix in suffixes]) + ")"
        rule = (regex, nearest)
        rules.append(rule)
    return rules


def imputer(tokens: list, rules: dict):
    for i, token in enumerate(tokens):
        for regex, nearest in rules:
            if re.match(regex, token):
                tokens[i] = nearest
    return tokens



from golem.core.chat_session import ChatSession, Profile

class TestInterface():
    name = 'test'
    prefix = 'test'
    messages = []
    states = []

    @staticmethod
    def clear():
        TestInterface.messages = []
        TestInterface.states = []

    @staticmethod
    def fill_session_profile(session: ChatSession):
        if not session:
            raise ValueError("Session is None")
        session.profile = Profile(first_name='Test', last_name='')
        return session

    @staticmethod
    def post_message(session, response):
        TestInterface.messages.append(response)

    @staticmethod
    def send_settings(settings):
        pass

    @staticmethod
    def processing_start(session):
        pass

    @staticmethod
    def processing_end(session):
        pass

    @staticmethod
    def state_change(state):
        if not TestInterface.states or TestInterface.states[-1] != state:
            TestInterface.states.append(state)

    @staticmethod
    def parse_message(user_message, num_tries=1):
        return user_message



import logging
import time
import traceback

from celery import shared_task
from celery.task.schedules import crontab
from celery.utils.log import get_task_logger
from django.conf import settings

from golem.core import message_logger  # this should register the celery log task
from golem.core.chat_session import ChatSession
from golem.core.interfaces.all import create_from_name
from golem.core.persistence import get_redis

logger = get_task_logger(__name__)


@shared_task
def accept_user_message(session: dict, raw_message):
    from golem.core.dialog_manager import DialogManager
    session = ChatSession.from_json(session)
    print("Accepting message - chat id {}, message: {}".format(session.chat_id, raw_message))

    dialog = DialogManager(session)
    parsed = session.interface.parse_message(raw_message)
    _process_message(dialog, parsed)

    should_log_messages = settings.GOLEM_CONFIG.get('SHOULD_LOG_MESSAGES', False)

    if should_log_messages and 'text' in raw_message:
        text = raw_message['text']
        message_logger.on_message.delay(session, text, dialog, from_user=True)
    return True  # FIXME


def setup_schedule_callbacks(sender, callback):
    callbacks = settings.GOLEM_CONFIG.get('SCHEDULE_CALLBACKS')
    if not callbacks:
        return

    for name in callbacks:
        params = callbacks[name]
        print('Scheduling task {}: {}'.format(name, params))
        if isinstance(params, dict):
            cron = crontab(**params)
        elif isinstance(params, int):
            cron = params
        else:
            raise Exception(
                'Specify either number of seconds or dict of celery crontab params (hour, minute): {}'.format(params))
        sender.add_periodic_task(
            cron,
            callback.s(name),
        )
        print(' Scheduled for {}'.format(cron))


def accept_schedule_all_users(callback_name):
    print('Accepting scheduled callback {}'.format(callback_name))
    db = get_redis()
    interface_names = db.hgetall('session_interface')
    for uid in interface_names:
        interface_name = interface_names[uid].decode('utf-8')
        # TODO revise this
        interface = create_from_name(interface_name)
        session = ChatSession(interface, uid.decode('utf-8'))
        accept_schedule_callback(session.to_json(), callback_name)


@shared_task
def accept_schedule_callback(session: dict, callback_name):
    session = ChatSession.from_json(session)
    from golem.core.dialog_manager import DialogManager
    db = get_redis()
    active_time = float(db.hget('session_active', session.chat_id).decode('utf-8'))
    inactive_seconds = time.time() - active_time
    print('{} from {} was active {}'.format(session.chat_id, session.interface, active_time))
    parsed = {
        'type': 'schedule',
        'entities': {
            'intent': '_schedule',
            '_inactive_seconds': inactive_seconds,
            '_callback_name': callback_name
        }
    }
    dialog = DialogManager(session)
    _process_message(dialog, parsed)


@shared_task
def accept_inactivity_callback(session: dict, context_counter, callback_name, inactive_seconds):
    session = ChatSession.from_json(session)
    from golem.core.dialog_manager import DialogManager
    dialog = DialogManager(session)

    # User has sent a message, cancel inactivity callback
    if dialog.context.counter != context_counter:
        print('Canceling inactivity callback after user message.')
        return

    parsed = {
        'type': 'schedule',
        'entities': {
            'intent': '_inactive',
            '_inactive_seconds': inactive_seconds,
            '_callback_name': callback_name
        }
    }

    _process_message(dialog, parsed)


def _process_message(dialog, parsed):
    try:
        dialog.process(parsed['type'], parsed['entities'])
    except Exception as e:
        print("!!!!!!!!!!!!!!!! EXCEPTION AT MESSAGE QUEUE !!!!!!!!!!!!!!!", e)
        traceback.print_exc()
        dialog.logger.log_error(exception=e, state=dialog.current_state_name)


@shared_task
def fake_move_to_state(chat_id, state: str, entities=()):
    # TODO
    session = get_redis().hget("chat_session", chat_id)

    if not (session and state):
        logging.warning("ChatSession or State is null")
        return

    import json
    session = json.loads(session.decode('utf-8'))
    session = ChatSession.from_json(session)
    state = str(state)
    from golem.core.dialog_manager import DialogManager
    logger.debug("Moving chat id {} to state {}".format(session.chat_id, state))
    dialog = DialogManager(session)
    msg_data = {'_state': state}
    for k, v in entities:
        msg_data[k] = [{"value": v}]
    dialog.process('postback', msg_data)



from abc import ABC, abstractmethod


class MessageLogger(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def log_user_message(self, dialog, time, state, message, type_, entities):
        pass

    @abstractmethod
    def log_bot_message(self, dialog, time, state, message):
        pass

    def log_error(self, dialog, state, exception):
        pass

    def log_user(self, dialog, chat_session):
        pass



import json

import os
from typing import List

import logging
from django.conf import settings

from golem.nlu import cleanup
from golem.nlu.data_model import Entity


def get_entity_names() -> List:
    training_dir = os.path.join(data_dir(), 'training_data')
    entities = []
    for f in os.scandir(training_dir):
        name, ext = os.path.splitext(f.name)
        if f.is_file() and ext == '.json':
            entities.append(Entity.fromFile(training_dir + '/' + f.name))
            # entities.append((name, f))
    return entities


_glove = None
_nlp = None


def get_glove():
    from golem.nlu.glove import GloVe
    global _glove
    if _glove is None:
        glove_dir = settings.NLP_CONFIG.get('GLOVE_DIR')
        glove_prefix = settings.NLP_CONFIG.get('GLOVE_PREFIX')
        if glove_dir is None or glove_prefix is None:
            raise Exception('Please set django settings NLP_CONFIG.GLOVE_DIR and GLOVE_PREFIX to point '
                            'to your word embedding files to use the built-in NLP.')
        logging.debug('Using word embeddings at {} with prefix'.format(glove_dir, glove_prefix))
        _glove = GloVe(glove_dir, glove_prefix)
    return _glove


def data_dir():
    return settings.NLP_CONFIG.get("DATA_DIR", "data/nlp")


def get_default_language():
    return settings.NLP_CONFIG.get('LANGUAGE', 'en')


def get_training_data(entity):
    path = os.path.join(data_dir(), "training_data", entity + ".json")
    with open(path) as f:
        data = json.load(f)
    return data


def get_imputation_rules(entity):
    data = get_training_data(entity)
    imputation = data.get('imputation', [])
    return cleanup.build_imputation_rules(imputation)



import time

from copy import deepcopy


# import sys
# import traceback

class Context:
    def __init__(self, dialog, entities, history, counter, max_depth=30, history_restart_minutes=30):
        self.counter = counter
        self.entities = entities
        self.history = history
        self.max_depth = max_depth
        self.dialog = dialog
        self.history_restart_minutes = history_restart_minutes

    def add_entities(self, new_entities):
        if new_entities is None:
            return
        inserted = {}
        # add all new entities
        for entity,values in new_entities.items():
            if entity not in self.entities:
                self.entities[entity] = []
            # allow also direct passing of {'entity' : 'value'}
            if not isinstance(values, dict) and not isinstance(values, list):
                values = {'value':values}
            if not isinstance(values, list):
                values = [values]
            
            inserted[entity] = deepcopy(values)
            # prepend each value to start of the list with 0 age 
            for value in values:
                value['counter'] = self.counter
                self.entities[entity] = [value] + self.entities[entity]
        self.debug()
        return inserted

    def add_state(self, state_name):
        timestamp = int(time.time())
        if len(self.history) > 20:
            self.history = self.history[-20:]

        state = {
            'name' : state_name,
            'timestamp' : timestamp
        }
        self.history.append(state)

    def clear(self, entities):
        for entity in entities:
            if entity in self.entities:
                del self.entities[entity]

    def get_min_entity_age(self, entities):
        ages = [self.get_age(entity)[1] for entity in entities]
        ages = filter(lambda x: x is not None, ages)
        return min(ages) if ages else None

    def get_history_state(self, index):
        return self.history[index] if len(self.history) > abs(index) else None

    def get_all(self, entity, max_age=None, limit=None, key='value', ignored_values=[]):
        values = []
        if entity not in self.entities:
            return values
        for value in self.entities[entity]:
            age = self.counter-value['counter']
            value['age'] = age
            # if I found a too old value, stop looking
            if max_age is not None and age > max_age:
                break
            v = value.get('value')
            if v in ignored_values:
                self.dialog.log.info('Skipping ignored entity value: {}={}'.format(entity, v))
                continue
            values.append(value.get(key) if key else value.copy())
            # if I already have enough values, stop looking
            if limit is not None and len(values) >= limit:
                break
        return values 
    
    def debug(self, max_age=5):
        self.dialog.log.info('-- HEAD of Context (max age {}): --'.format(max_age))
        for entity in self.entities:
            values = self.get_all_first(entity, key=None, max_age=max_age)
            if values:
                vs = [value.get('value') for value in values]
                self.dialog.log.info('{} (age {}): {}'.format(entity, values[0]['age'], vs if len(vs) > 1 else vs[0]))
        self.dialog.log.info('----------------------------------')

    def get(self, entity, max_age=None, key='value', ignored_values=[]):
        values = self.get_all(entity, max_age=max_age, limit=1, key=key, ignored_values=ignored_values)
        if not values:
            return None
        return values[0]

    def get_age(self, entity, max_age=None, key='value', ignored_values=[]):
        value = self.get_all(entity, max_age=max_age, limit=1, key=None, ignored_values=ignored_values)
        if not value:
            return (None, None)
        return value[0].get(key) if key else value[0], value[0]['age'] 

    def get_all_first(self, entity, max_age=None, key='value'):
        values = []
        if entity not in self.entities:
            return values
        found_age = None
        existing = []
        for value in self.entities[entity]:
            age = self.counter-value['counter']
            value['age'] = age
            # if I found a too old value, stop looking
            if max_age is not None and age > max_age:
                break
            if found_age is not None and age > found_age:
                break
            found_age = age
            if value.get('value') in existing:
                # skip duplicates
                continue
            existing.append(value.get('value'))
            values.append(value.get(key) if key else value)
        return values[::-1]

    def set(self, entity, value_dict):
        if not isinstance(value_dict, dict):
            raise ValueError('Use a dict to set a context value, e.g. {"value":"foo"}. Call multiple times to add more.')
        if entity not in self.entities:
            self.entities[entity] = []
        value_dict['counter'] = self.counter
        self.entities[entity] = [value_dict] + self.entities[entity][:self.max_depth-1]

    def set_value(self, entity, value):
        if entity not in self.entities:
            self.entities[entity] = []
        value_dict = {'counter': self.counter, 'value': value}
        self.entities[entity] = [value_dict] + self.entities[entity][:self.max_depth-1]

    def has_any(self, entities, max_age=None):
        for entity in entities:
            if self.get(entity, max_age=max_age, key=None):
                return True
        return False

    def has_all(self, entities, max_age=None):
        for entity in entities:
            if not self.get(entity, max_age=max_age, key=None):
                return False
        return True

    #def entity_value(self, entity):
    #    values = self.entity_values(entity)
    #    return values[0] if values else None

    #def __str__(self):
    #    res = "Message: %s\n" % (self.text)
    #    for entity in self.entities:
    #        for value in self.entities[entity]:
    #            res += "Entity %s: %s (%s)\n" % (entity, value['value'], value['confidence'])


import json
import logging
import pickle

from wit import Wit

from golem.core.parsing import date_utils
from golem.core.parsing.entity_extractor import EntityExtractor
from golem.core.persistence import get_redis


class WitExtractor(EntityExtractor):

    def __init__(self, wit_token, enable_cache=True):
        super().__init__()
        self.log = logging.getLogger()
        self.wit_token = wit_token
        if not self.wit_token:
            raise ValueError("Wit token not found!")
        self.cache_key = 'wit_cache'
        self.cache = enable_cache
        # self.clear_wit_cache()

    def extract_entities(self, text: str, max_retries=5):
        if max_retries <= 0:
            self.log.error("Maximal number of Wit retries reached")
            return {}
        cached = self._load_from_cache(text)
        if cached: return cached
        try:
            wit_client = Wit(access_token=self.wit_token, actions={})
            entities = wit_client.message(text).get('entities', {})
            entities = self._process_wit_entities(entities)
            self.save_to_cache(text, entities)
            return entities
        except Exception as e:
            self.log.exception('Wit error:', e)
            return self.extract_entities(text, max_retries - 1)

    def _process_wit_entities(self, entities: dict):

        entities = self._process_metadata(entities)

        if 'datetime' in entities:
            datetime = entities['datetime']
            duration = entities.get('duration', None)
            append = date_utils.process_datetime(datetime, duration)
            entities.update(append)

        return entities

    def _process_metadata(self, entities: dict):
        for entity, values in entities.items():
            for value in values:
                # parse string metadata from Wit into a dict
                metadata = value.get('metadata')
                if metadata and isinstance(metadata, str):
                    try:
                        value['metadata'] = json.loads(metadata)
                    except:
                        self.log.warning("Ignoring invalid metadata for entity {}: {}".format(
                            entity, metadata
                        ))
                        value['metadata'] = None
        return entities

    def _load_from_cache(self, text):
        if self.cache:
            db = get_redis()
            if db.hexists('wit_cache', text):
                parsed = pickle.loads(db.hget('wit_cache', text))
                self.log.debug('Got cached wit key: "{}" = {}'.format(self.cache_key, parsed))
                return parsed
        return None

    def save_to_cache(self, text, entities):
        if self.cache and 'date_interval' not in entities:
            db = get_redis()
            self.log.debug('Caching wit key: {} = {}'.format(text, entities))
            db.hset('wit_cache', text, pickle.dumps(entities))

    def clear_wit_cache(self):
        if self.cache:
            self.log.debug('Clearing Wit cache...')
            db = get_redis()
            db.delete('wit_cache')

def teach_wit(wit_token, entity, values, doc=""):
    import requests
    print('*** TEACHING WIT ***')
    params = {'v':'20160526'}
    print('Inserting values of {}'.format(entity))
    rsp = requests.request(
        'PUT',
        'https://api.wit.ai/entities/'+entity,
        headers={
            'authorization': 'Bearer ' + wit_token,
            'accept': 'application/json'
        },
        params=params,
        json={'doc':doc or entity, 'values':values}
    )
    if rsp.status_code > 200:
        raise ValueError('Wit responded with status: ' + str(rsp.status_code) +
                       ' (' + rsp.reason + '): ' + rsp.text)
    json = rsp.json()
    if 'error' in json:
        raise ValueError('Wit responded with an error: ' + json['error'])



from django.conf import settings
from django.conf.urls import url

from . import views

secret_url = settings.GOLEM_CONFIG.get('WEBHOOK_SECRET_URL')
urlpatterns = [
    url(r'^messenger/{}/?$'.format(secret_url),
        view=views.FacebookView.as_view(), name='messenger'),
    url(r'^telegram/{}/?$'.format(secret_url),
        view=views.TelegramView.as_view(), name='telegram'),
    url(r'^gactions/{}/?$'.format(secret_url),
        view=views.GActionsView.as_view(), name='gactions'),
    url(r'^skype/{}/?$'.format(secret_url),
        view=views.SkypeView.as_view(), name='skype'),
    url(r'^run_all_tests/?$', views.run_all_tests),
    url(r'^run_test/(?P<name>[a-zA-Z0-9_\-]+)/?$', views.run_test),
    url(r'^run_test_message/(?P<message>[a-zA-Z0-9 _\-]+)/?$', views.run_test_message),
    url(r'^log/(?P<user_limit>[0-9]*)/?$', views.log),
    url(r'^log_tests/?$', views.log_tests),
    url(r'^test/?$', views.test),
    url(r'^debug/?$', views.debug),
    url(r'^test_record/?$', views.test_record),
    url(r'^users/?', views.users_view),
    url(r'^log_conversation/(?P<group_id>[a-zA-Z_0-9]*)/(?P<page>[0-9]*)/?$', views.log_conversation)
]



# TODO https://stackoverflow.com/questions/10572603/specifying-optional-dependencies-in-pypi-python-setup-py
import pickle

import numpy as np
import os
import random
import yaml

from golem.nlu import cleanup
from golem.nlu import utils
from golem.nlu.keywords import TrieKeywordModel
from golem.nlu.nn.bow_model import BowModel
from golem.nlu.nn.contextual import ContextualModel
from golem.nlu.nn.rnn_intent_model import RecurrentIntentModel


def process(entities, imputation_rules):
    """
    Processes entities to stemmed words with SpaCy.
    :returns:   tuple of words, documents, classes
    """

    ignore_words = ['?', '.', '!']
    words = []
    documents = []
    classes = []

    for entity in entities:
        # specific entity value with samples
        value = entity['value']

        if value not in classes:
            classes.append(value)

        for sample in entity['samples']:
            # a text pattern
            tokens = cleanup.tokenize(sample)
            tokens = cleanup.imputer(tokens, imputation_rules)
            words.extend(tokens)
            documents.append((tokens, value))

    words = sorted(list(set(words)))
    words = [w for w in words if w not in ignore_words]

    print(len(documents), 'documents')
    print(len(classes), 'classes')
    print(len(words), 'words', words)

    classes.append('none')
    documents.append(([], 'none'))  # empty sentence to prevent bias @ [0]

    return words, documents, classes


class Batcher:
    def __init__(self, training, max_words=10, use_vocab=False):
        self.glove = utils.get_glove()
        self.dim = self.glove.get_dimension()
        self.max_words = max_words
        self.imputation_rules = cleanup.build_imputation_rules(training.get('imputation', []))
        data = training['data']
        self.keep_prob = training.get("dropout", 0.5)
        self.iterations = training.get("iterations", 3000)
        self.labels = [x['value'] for x in data]  # TODO what if some are duplicates?
        self.sentences = dict((x['value'], x['samples']) for x in data)
        self.sentences.setdefault(None, []).append([])
        self.unk = self.glove.get_vector("<unk>")
        if self.unk is None: self.unk = np.zeros([self.dim])
        self.oov = set()

        if use_vocab:
            self.vocab = self.make_vocab(self.sentences.values())

            self.sentences_bow = dict(  # converts sentences to bag of words features
                (k, list(filter(None, [self.encode_sentence_bow(sent) for sent in v])))
                for k, v in self.sentences.items()
            )

        self.sentences = dict(  # converts sentences to embedded feature vectors
            (k, list(filter(None, [self.process_sentence(sent) for sent in v])))
            for k, v in self.sentences.items()
        )

        # TODO somehow distinguish PAD, END, START, UNK and NIL

        if len(self.oov) > 0:
            print("### [Warning] There are %d out-of-vocabulary words in training data! ###" % len(self.oov))
            print("OOV words: ", sorted(self.oov))

    def process_sentence(self, sentence):
        tokens = cleanup.imputer(cleanup.tokenize(sentence), self.imputation_rules)
        features = np.zeros([self.max_words, self.dim])
        random_positions, has_valid_token = [], False
        for i, token in enumerate(tokens):
            if token == "%":
                random_positions.append(i)
                has_valid_token = True  # well well, but it might get kinda fuzzy ...
            else:
                vec = self.glove.get_vector(token)
                if vec is None:
                    self.oov.add(token)
                    features[i] = self.unk
                else:
                    has_valid_token = True
                    features[i] = vec
        if not has_valid_token:
            print("### [Warning] Sentence \"{}\" has no valid words! Removing! ###".format(sentence))
        return (features, random_positions) if has_valid_token else None

    def make_vocab(self, sentences):
        vocab = set()
        for sent in sentences:
            tokens = cleanup.imputer(cleanup.tokenize(sent), self.imputation_rules)
            vocab += tokens
        return list(vocab)

    def encode_sentence_bow(self, sentence):
        bow = [0.] * len(self.vocab)
        tokens = cleanup.imputer(cleanup.tokenize(sentence), self.imputation_rules)
        for token in tokens:
            if token in self.vocab:
                bow[self.vocab.index(token)] = 1.0
        return bow

    def next_batch(self, batch_size):
        # ensure proper stratification
        labels = [random.choice(self.labels) for _ in range(batch_size)]
        x, y = [], []
        for i, label in enumerate(labels):
            sentence, positions = random.choice(self.sentences[label])
            for pos in positions:
                sentence[pos] = np.random.random([self.dim])
            x.append(sentence)
            one_hot = [0.] * len(self.labels)
            if label not in [None, 'none', 'None']:
                one_hot[self.labels.index(label)] = 1.0
            y.append(one_hot)
        return x, y

    def next_batch_bow(self, batch_size):
        labels = [random.choice(self.labels) for _ in range(batch_size)]
        x, y = [], []
        for i, label in enumerate(labels):
            sentence = random.choice(self.sentences_bow[label])
            x.append(sentence)
            one_hot = [0.] * len(self.labels)
            if label not in [None, 'none', 'None']:
                one_hot[self.labels.index(label)] = 1.0
            y.append(one_hot)
        return x, y


def train_entity(batcher, entity_name, entity_dir, num_iterations):
    """
    Trains a model for recognizing entity based on x, y.
    """
    print("[Training entity {} for {} iterations]".format(entity_name, num_iterations))
    model = RecurrentIntentModel(entity_name, entity_dir)
    model.train(batcher, batcher.iterations, batcher.keep_prob)
    model.destroy()


def trashify(x):
    """Replace all wildcards in text by random trash words."""
    glove = utils.get_glove()
    while '%' in x:
        junk = None
        while not junk or '%' in junk:
            junk = glove.random_word()
        x = x.replace('%', junk, 1)
    return x


def train_fuzzy_matcher(examples):
    pass


def train_all(included=None):
    """
    Trains all entity values from their JSON descriptions.
    See ./data/training_data/*.json
    """
    train_dir = os.path.join(utils.data_dir(), 'training_data')
    entities = []
    for f in os.scandir(train_dir):
        name, ext = os.path.splitext(f.name)
        if f.is_file() and ext in ['.json', '.yaml', '.yml']:
            entities.append((name, f))

    for entity, filename in entities:
        if included and entity not in included:
            continue
        print('Training', entity)
        with open(filename) as f:
            data = yaml.load(f)
            entity_dir = os.path.join(utils.data_dir(), 'model', entity)
            if not os.path.exists(entity_dir):
                os.makedirs(entity_dir)

            strategy = data['strategy']
            with open(os.path.join(entity_dir, 'metadata.json'), 'w') as g:
                metadata = {'strategy': strategy}
                if strategy == 'trait':
                    metadata['threshold'] = data.get('threshold', 0.5)
                if strategy == 'keywords':
                    # metadata['ngrams'] = data.get('ngrams', 3)
                    metadata['stemming'] = data.get('stemming', False)
                    metadata['language'] = data.get('language', utils.get_default_language())
                yaml.dump(metadata, g)

            if strategy == "bow":
                model = BowModel(entity, entity_dir, is_training=True)
                model.train(data['data'])
            elif strategy == 'context':
                model = ContextualModel(entity, entity_dir, is_training=True)
                model.train(data['data'])
            elif strategy == 'trait':
                # train as neural network
                # samples = data['data']
                # imputation = data.get('imputation', [])
                num_iterations = data.get("iterations", 1000)
                # words, documents, classes = process(samples, rules)
                words, documents, classes = [], [], []
                entity_dir = os.path.join(utils.data_dir(), 'model', entity)
                batcher = Batcher(data, max_words=10)
                train_entity(batcher, entity, entity_dir, num_iterations)
                pickle_path = os.path.join(utils.data_dir(), 'model', entity, 'metadata.pkl')
                pickle.dump({'labels': batcher.labels},
                            open(pickle_path, 'wb'))
            elif strategy == 'keywords':
                # train as a list of fixed values (fuzzy matching)
                samples = data['data']
                should_stem = data.get('stemming', False)
                language = data.get('language', utils.get_default_language())

                TrieKeywordModel(entity, entity_dir, is_training=True).prepare(
                    examples=samples,
                    metadata={"stemming": should_stem, "language": language}
                )

            elif strategy == 'fuzzy':
                samples = data['data']
                train_fuzzy_matcher(samples)
            else:
                print("Unknown training strategy {} for entity {}, skipping!".format(strategy, entity))

    print("All entities trained!")



#! /usr/bin/env python3.1
""" Czech stemmer
Copyright © 2010 Luís Gomes <luismsgomes@gmail.com>.
This code is released under a Creative Commons Attribution 3.0 Unported License.
https://creativecommons.org/licenses/by/3.0/

Ported from the Java implementation available at:
    http://members.unine.ch/jacques.savoy/clef/index.html

"""
import re
import sys


### FIXME VALIDATE SLOVAK CHARACTERS !!! ###

def cz_stem_all(words, aggressive=True):
    return [cz_stem(word, aggressive) for word in words]


def cz_stem(word, aggressive=False):
    if not re.match("^\\w+$", word):
        return word
    if not word.islower() and not word.istitle() and not word.isupper():
        print("warning: skipping word with mixed case: {}".format(word),
              file=sys.stderr)
        return word
    s = word.lower()  # all our pattern matching is done in lowercase
    s = _remove_case(s)
    s = _remove_possessives(s)
    if aggressive:
        s = _remove_comparative(s)
        s = _remove_diminutive(s)
        s = _remove_augmentative(s)
        s = _remove_derivational(s)
    if word.isupper():
        return s.upper()
    if word.istitle():
        return s.title()
    return s


def _remove_case(word):
    if len(word) > 7 and word.endswith("atech"):
        return word[:-5]
    if len(word) > 6:
        if word.endswith("ětem"):
            return _palatalise(word[:-3])
        if word.endswith("atům"):
            return word[:-4]
    if len(word) > 5:
        if word[-3:] in {"ech", "ich", "ích", "ého", "ěmi", "emi", "ému",
                         "ete", "eti", "iho", "ího", "ími", "imu"}:
            return _palatalise(word[:-2])
        if word[-3:] in {"ách", "ata", "aty", "ých", "ama", "ami",
                         "ové", "ovi", "ými"}:
            return word[:-3]
    if len(word) > 4:
        if word.endswith("em"):
            return _palatalise(word[:-1])
        if word[-2:] in {"es", "ém", "ím"}:
            return _palatalise(word[:-2])
        if word[-2:] in {"ům", "at", "ám", "os", "us", "ým", "mi", "ou"}:
            return word[:-2]
    if len(word) > 3:
        if word[-1] in "eiíě":
            return _palatalise(word)
        if word[-1] in "uyůaoáéý":
            return word[:-1]
    return word


def _remove_possessives(word):
    if len(word) > 5:
        if word[-2:] in {"ov", "ův"}:
            return word[:-2]
        if word.endswith("in"):
            return _palatalise(word[:-1])
    return word


def _remove_comparative(word):
    if len(word) > 5:
        if word[-3:] in {"ejš", "ějš"}:
            return _palatalise(word[:-2])
    return word


def _remove_diminutive(word):
    if len(word) > 7 and word.endswith("oušek"):
        return word[:-5]
    if len(word) > 6:
        if word[-4:] in {"eček", "éček", "iček", "íček", "enek", "ének",
                         "inek", "ínek"}:
            return _palatalise(word[:-3])
        if word[-4:] in {"áček", "aček", "oček", "uček", "anek", "onek",
                         "unek", "ánek"}:
            return _palatalise(word[:-4])
    if len(word) > 5:
        if word[-3:] in {"ečk", "éčk", "ičk", "íčk", "enk", "énk",
                         "ink", "ínk"}:
            return _palatalise(word[:-3])
        if word[-3:] in {"áčk", "ačk", "očk", "učk", "ank", "onk",
                         "unk", "átk", "ánk", "ušk"}:
            return word[:-3]
    if len(word) > 4:
        if word[-2:] in {"ek", "ék", "ík", "ik"}:
            return _palatalise(word[:-1])
        if word[-2:] in {"ák", "ak", "ok", "uk"}:
            return word[:-1]
    if len(word) > 3 and word[-1] == "k":
        return word[:-1]
    return word


def _remove_augmentative(word):
    if len(word) > 6 and word.endswith("ajzn"):
        return word[:-4]
    if len(word) > 5 and word[-3:] in {"izn", "isk"}:
        return _palatalise(word[:-2])
    if len(word) > 4 and word.endswith("ák"):
        return word[:-2]
    return word


def _remove_derivational(word):
    if len(word) > 8 and word.endswith("obinec"):
        return word[:-6]
    if len(word) > 7:
        if word.endswith("ionář"):
            return _palatalise(word[:-4])
        if word[-5:] in {"ovisk", "ovstv", "ovišt", "ovník"}:
            return word[:-5]
    if len(word) > 6:
        if word[-4:] in {"ásek", "loun", "nost", "teln", "ovec", "ovík",
                         "ovtv", "ovin", "štin"}:
            return word[:-4]
        if word[-4:] in {"enic", "inec", "itel"}:
            return _palatalise(word[:-3])
    if len(word) > 5:
        if word.endswith("árn"):
            return word[:-3]
        if word[-3:] in {"ěnk", "ián", "ist", "isk", "išt", "itb", "írn"}:
            return _palatalise(word[:-2])
        if word[-3:] in {"och", "ost", "ovn", "oun", "out", "ouš",
                         "ušk", "kyn", "čan", "kář", "néř", "ník",
                         "ctv", "stv"}:
            return word[:-3]
    if len(word) > 4:
        if word[-2:] in {"áč", "ač", "án", "an", "ář", "as"}:
            return word[:-2]
        if word[-2:] in {"ec", "en", "ěn", "éř", "íř", "ic", "in", "ín",
                         "it", "iv"}:
            return _palatalise(word[:-1])
        if word[-2:] in {"ob", "ot", "ov", "oň", "ul", "yn", "čk", "čn",
                         "dl", "nk", "tv", "tk", "vk"}:
            return word[:-2]
    if len(word) > 3 and word[-1] in "cčklnt":
        return word[:-1]
    return word


def _palatalise(word):
    if word[-2:] in {"ci", "ce", "či", "če"}:
        return word[:-2] + "k"

    if word[-2:] in {"zi", "ze", "ži", "že"}:
        return word[:-2] + "h"

    if word[-3:] in {"čtě", "čti", "čtí"}:
        return word[:-3] + "ck"

    if word[-3:] in {"ště", "šti", "ští"}:
        return word[:-3] + "sk"
    return word[:-1]



import json
import os
import string

import random
import tensorflow as tf
import numpy as np


class Seq2Seq:
    def __init__(self, entity, base_dir):
        if not base_dir or not entity:
            raise ValueError('Base dir and entity name must be set')
        self.base_dir = base_dir
        self.entity = entity
        self.graph = tf.Graph()
        self.session = tf.Session(graph=self.graph)
        self.loaded = False
        self.label_cnt = None
        self.alphabet = list(string.ascii_lowercase + string.digits + ' ')

    def destroy(self):
        self.session.close()

    def make_weights(self, shape):
        return tf.Variable(tf.truncated_normal(shape, stddev=0.1))

    def make_biases(self, shape):
        return tf.Variable(tf.constant(0.1, shape=shape))

    def loss_fn(self, labels, logits):
        return tf.contrib.seq2seq.sequence_loss(
            logits=logits,
            targets=tf.cast(tf.argmax(labels, axis=2), tf.int32),
            weights=tf.ones([logits.shape[0], logits.shape[1]])
        )

    def get_word_lengths(self, batch):
        raise Exception('dont use me')
        nonzero = tf.sign(tf.reduce_max(tf.abs(batch), axis=2))  # number of filled features
        length = tf.cast(tf.reduce_sum(nonzero, axis=1), tf.int32)  # as int
        return length

    def get_last_component(self, lstm_output, length):
        batch_size = tf.shape(lstm_output)[0]
        max_length = tf.shape(lstm_output)[1]
        out_size = int(lstm_output.get_shape()[2])
        index = tf.range(0, batch_size) * max_length + (length - 1)
        flat = tf.reshape(lstm_output, [-1, out_size])
        relevant = tf.gather(flat, index)
        return relevant

    def encode_to_alphabet(self, text, length, alphabet) -> np.array:
        alphabet_size = len(alphabet)
        text = text[0:length].lower()
        encoded_chars = [np.zeros(alphabet_size) for i in range(length)]
        for i in range(length):
            encoded_chars[i][alphabet.index(' ')] = 1.0
        for char_off, char in enumerate(text):
            if char in alphabet:
                idx = alphabet.index(char)
                encoded = np.zeros(alphabet_size)
                encoded[idx] = 1.0
                encoded_chars[char_off] = encoded
                # encoded_chars[char_off][idx] = 1.0
        return np.array(encoded_chars)

    def decode_from_alphabet(self, characters, alphabet) -> str:
        text = ""
        for char in characters:
            val = np.max(char)
            if val > 0.8:
                idx = np.argmax(char)
                text += alphabet[idx]
            else:
                text += '_'
        return text

    def get_shape(self, max_length, alphabet_size, batch_size=1):
        memory_size = int(alphabet_size * 2)

        x = tf.placeholder(tf.float32, shape=[batch_size, max_length, alphabet_size], name='x')
        y = tf.placeholder(tf.float32, shape=[batch_size, max_length, alphabet_size], name='y')

        with tf.variable_scope('encoder') as scope:
            encoder_out, state = tf.nn.dynamic_rnn(
                tf.contrib.rnn.LSTMCell(alphabet_size * 5, forget_bias=0.9),
                x,
                scope=scope,
                dtype=tf.float32
            )

        with tf.variable_scope('decoder') as scope:
            decoder_out, state = tf.nn.dynamic_rnn(
                tf.contrib.rnn.LSTMCell(alphabet_size, forget_bias=0.9, activation=tf.nn.relu),
                encoder_out,
                scope=scope,
                dtype=tf.float32
            )

        y_ = decoder_out

        return x, y, y_

    def randomize(self, features, labels):
        x_out, y_out = [], []
        for x, y in zip(features, labels):
            length = random.randint(1, 10)
            s = ''.join([random.choice(string.ascii_lowercase + string.digits + ' ') for i in range(length)])
            x_out.append(x.replace('%', s))
            y_out.append(y.replace('%', s))
        return x_out, y_out

    def train(self, features, labels, num_epochs=50000, max_length=25):  # FIXME more than 25 chars
        if features is None or labels is None:
            raise ValueError('Features and labels must be set')

        features = np.array(features)
        labels = np.array(labels)

        with self.graph.as_default():

            batch_size = 1

            x, y, y_ = self.get_shape(max_length, len(self.alphabet), batch_size=batch_size)

            loss = self.loss_fn(y, y_)

            sample_cnt = features.shape[0]

            optimizer = tf.train.AdamOptimizer(0.001).minimize(loss)
            accuracy = tf.metrics.accuracy(tf.argmax(tf.nn.softmax(y_), 1), tf.argmax(y, 1))

            self.session.run(tf.global_variables_initializer())
            self.session.run(tf.local_variables_initializer())
            for i in range(num_epochs):
                start_idx = i % (sample_cnt - batch_size)
                end_idx = start_idx + batch_size

                _x_, _y_ = self.randomize(features[start_idx:end_idx], labels[start_idx:end_idx])
                _x_ = [self.encode_to_alphabet(x, max_length, self.alphabet) for x in _x_]
                _y_ = [self.encode_to_alphabet(y, max_length, self.alphabet) for y in _y_]

                optimizer.run(feed_dict={x: _x_, y: _y_},
                              session=self.session)
                print('Generation ', i, '/', num_epochs,
                      'Xentropy:', self.session.run(loss, feed_dict={x: _x_, y: _y_}),
                      'Accuracy:', self.session.run(accuracy, feed_dict={x: _x_, y: _y_}))
                output = self.session.run(y_, feed_dict={x: _x_, y: _y_})
                print(
                    self.decode_from_alphabet(output[0], self.alphabet)
                    + "\t"
                    + self.decode_from_alphabet(_y_[0], self.alphabet)
                )  # it's a batch

            # save model
            saver = tf.train.Saver()
            saver.save(self.session, os.path.join(self.base_dir, self.entity, 'tf_session'))
            with open(os.path.join(self.base_dir, self.entity, 'tf_metadata.json'), 'w') as f:
                json.dump({'max_length': max_length}, f)

    def load(self):
        with open(os.path.join(self.base_dir, self.entity, 'tf_metadata.json')) as f:
            metadata = json.load(f)
        # feature_cnt, label_cnt = metadata['feature_cnt'], metadata['label_cnt']
        # word_cnt = metadata['word_cnt']
        # print('features:', feature_cnt, 'labels:', label_cnt)
        print('Restoring from {}'.format(self.base_dir))
        x, y, y_ = self.get_shape(25, len(self.alphabet))  # FIXME
        self.x = x
        self.y = y
        self.logits = y_
        self.y_ = tf.nn.softmax(y_)
        # self.W = W
        # self.label_cnt = label_cnt
        saver = tf.train.Saver()
        saver.restore(self.session, os.path.join(self.base_dir, self.entity, 'tf_session'))

    def predict(self, text):
        with self.graph.as_default():
            if not self.loaded:
                self.load()
                self.loaded = True

            _x_ = self.encode_to_alphabet(text, 25, self.alphabet)  # FIXME max_length
            pred = self.session.run(self.logits, feed_dict={self.x: [_x_]})
            s = self.decode_from_alphabet(pred[0], self.alphabet)
            print(s)
            return s.lstrip().rstrip()

    def test(self, x, y):
        raise NotImplementedError()
        with self.graph.as_default():
            if not self.loaded:
                self.load()
                self.loaded = True

            # accuracy = tf.metrics.accuracy(predictions=tf.argmax(self.y_, 1), labels=tf.argmax(self.y, 1))
            scores, predictions, losses = [], [], []
            loss_fn = self.loss_fn(self.y, self.logits, self.W)
            for _x_, _y_ in zip(x, y):
                labels = np.zeros(self.label_cnt)
                labels[_y_] = 1.0

                y_ = self.session.run(self.y_, feed_dict={self.x: _x_})
                loss = self.session.run(loss_fn, feed_dict={self.x: _x_, self.y: [labels]})
                y_ = np.argmax(y_)
                predictions.append(y_)
                losses.append(loss)
                if y_ == _y_:
                    scores.append(1.0)
                else:
                    scores.append(0.0)
            print("[{}] Accuracy score: {}% or {}/{}".format(self.entity, np.mean(scores) * 100, scores.count(1.0),
                                                             len(scores)))
            print("[{}] Mean Cross Entropy Loss: {}".format(self.entity, np.mean(losses)))
            return scores, predictions



from django.db import models


class User(models.Model):
    uid = models.CharField(max_length=255, primary_key=True)
    first_name = models.CharField(max_length=255, blank=True, null=True)
    surname = models.CharField(max_length=255, blank=True, null=True)
    message_cnt = models.PositiveIntegerField(default=0)

    def get_chats(self):
        chats = Chat.objects.filter(user_uid=self)
        return chats


class Chat(models.Model):
    chat_id = models.CharField(max_length=255, primary_key=True)
    user_uid = models.ForeignKey(User, on_delete=models.CASCADE)


class Message(models.Model):
    id = models.BigAutoField(primary_key=True)
    chat = models.ForeignKey(Chat, on_delete=models.CASCADE)
    text = models.TextField(blank=True, null=True)
    is_from_user = models.BooleanField()
    time = models.DateTimeField(auto_now=True)
    intent = models.TextField(max_length=255, blank=True, null=True, db_index=True)
    state = models.TextField(max_length=255, blank=True, null=True, db_index=True)


class QuickReply(models.Model):
    message = models.ForeignKey(Message, on_delete=models.CASCADE)
    text = models.TextField(max_length=255, blank=False, null=False)
    # TODO was_clicked : boolean

# class Attachment(models.Model):  TODO
#     message = models.ForeignKey(Message, on_delete=models.CASCADE)

# class ImageAttachment(Attachment):
#     image_url = models.TextField(max_length=4096, blank=True, null=True)



from django import template

register = template.Library()

@register.filter(name='get')
def get(d, k):
    return d.get(k, None)

@register.filter(name='entity')
def entity(d, k):
    v = d.get(k, None) if d else None
    return v[0].get('value') if v and len(v) else None

@register.filter(name='duration')
def duration(seconds):
    if not seconds:
      return ''
    return '{} ms'.format(int(seconds * 1000))





class Templates:
    @staticmethod
    def message(message, next=None):
        def action(state):
            return message, next
        return action
    
    @staticmethod
    def input(entity, missing_message=None, accept_text=False, next=None):
        def action(state):
            if accept_text:
                state.dialog.context.set(entity, {"value":state.dialog.context.get('_message_text')})
            elif not state.dialog.context.get(entity, max_age=0):
                return missing_message, None
            return None, next
        return action

    @staticmethod
    def value_transition(entity, transitions, next=None):
        def action(state):
            value = state.dialog.context.get(entity)
            if value in transitions.keys():
                return None, transitions[value]
            return None, next
        return action

        


from typing import List

from golem.core.responses.responses import MessageElement, GenericTemplateElement


class ListTemplate(MessageElement):

    def __init__(self, is_compact=False, button=None):
        super().__init__()
        self.elements = []  # type: List[GenericTemplateElement]
        self.compact = is_compact
        self.button = button

    def add_element(self, element: GenericTemplateElement):
        self.elements.append(element)
        return self

    def create_element(self, **kwargs):
        element = GenericTemplateElement(**kwargs)
        return self.add_element(element)

    def to_response(self):
        response = {
            "attachment": {
                "type": "template",
                "payload": {
                    "template_type": "list",
                    'top_element_style': 'compact' if self.compact else 'large',
                    "elements": [element.to_response() for element in self.elements[:4]]
                }
            }
        }
        if self.button:
            response['attachment']['payload']['buttons'] = [self.button.to_response()]

        return response



import json
from datetime import datetime

from django.conf import settings
from django.db.models import Max
from django.http import HttpResponse, HttpResponseBadRequest, JsonResponse
from django.shortcuts import render, redirect

from golem.core.responses import TextMessage
from .forms import MessageForm
from .interface import WebGuiInterface
from .models import WebMessageData


def webgui(request):
    if 'uid' not in request.session:
        return render(request, 'webchat/welcome.html')

    uid = request.session['uid']

    if request.method == 'POST':
        # message received via webgui
        if request.POST.get('message'):
            msg = WebMessageData()
            msg.uid = uid
            message_text = request.POST.get('message')
            msg.is_response = False
            msg.data = json.dumps(TextMessage(message_text).to_response())

            if message_text:
                # process message if text != null
                msg.save()
                if request.POST.get("postback"):
                    postback = request.POST.get("postback")
                    WebGuiInterface.accept_postback(msg, postback)
                else:
                    WebGuiInterface.accept_request(msg)
        else:
            print("Error, message not set in POST")
            return HttpResponseBadRequest()
        return HttpResponse()

    messages = WebMessageData.objects.filter(uid=uid).order_by('timestamp')
    context = {
        'uid': uid, 'messages': messages,
        'form': MessageForm, 'timestamp': datetime.now().timestamp(),
        'user_img': settings.GOLEM_CONFIG.get('WEBGUI_USER_IMAGE', 'images/icon_user.png'),
        'bot_img': settings.GOLEM_CONFIG.get('WEBGUI_BOT_IMAGE', 'images/icon_robot.png')
    }
    return render(request, 'webchat/index.html', context)


def do_login(request):
    if request.method == 'POST' and 'username' in request.POST:
        username = request.POST.get('username')
        uid = WebGuiInterface.make_uid(username)
        request.session['uid'] = uid
        request.session['username'] = username
        return HttpResponse()
    else:
        return HttpResponseBadRequest()


def do_logout(request):
    if 'uid' in request.session:
        WebGuiInterface.destroy_uid(request.session['uid'])
        del request.session['uid']
        del request.session['username']
    return redirect('webgui')


def get_last_change(request):
    if 'uid' in request.session:
        uid = request.session['uid']
        max_timestamp = WebMessageData.objects.filter(uid=uid).aggregate(Max('timestamp'))
        return JsonResponse(max_timestamp)
    return HttpResponseBadRequest()



import json
import logging
import os
from typing import Optional
from urllib.request import Request, urlopen

import requests

GENEEA_HEADERS = {
    'content-type': 'application/json',
    'Authorization': 'user_key ' + os.environ.get('GENEEA_API_KEY', "")
}


def extract_tags(title: Optional[str], text: Optional[str]):
    """
    Extracts tags from text using Geneea web service.
    """
    if 'GENEEA_API_KEY' not in os.environ:
        return []

    message = {
        'returnTextInfo': False,
        "domain": os.environ.get("GENEEA_DOMAIN"),
        "correction": "aggressive",
        "diacritization": "yes"
    }
    if title:
        message['title'] = title.strip()
    if text:
        message['text'] = text.strip()
    content = json.dumps(message).encode('utf-8')
    req = Request('https://api.geneea.com/s2/tags', headers=GENEEA_HEADERS, data=content)
    resp_conn = urlopen(req)
    response = json.loads(resp_conn.read().decode('utf-8'))
    return response['tags']


def get_sentiment(text: str, full=False):
    """
    Gets sentiment (positive/negative) of a text using Geneea web service.
    """
    if 'GENEEA_API_KEY' not in os.environ:
        return None

    if type(text) is not str:
        raise Exception('Illegal argument, text not str')
    message = {'returnTextInfo': False, 'text': text}
    content = json.dumps(message).encode('utf8')
    req = Request('https://api.geneea.com/s2/sentiment', headers=GENEEA_HEADERS, data=content)
    conn = urlopen(req)
    resp = json.loads(conn.read().decode('utf8'))
    return resp if full else float(resp['sentiment'])


def get_correction(text: str):
    if 'GENEEA_API_KEY' not in os.environ:
        return text

    data = {
        "text": text,
        "correction": "basic",
        "diacritization": "auto",
        "language": "cs"
    }
    req = requests.post(url="https://api.geneea.com/s2/correction", headers=GENEEA_HEADERS, data=json.dumps(data))
    if req.ok:
        return req.json().get("text", "")
    logging.warning("Geneea request for correction failed with code {} reason {}".format(req.status_code, req.reason))
    return text



import logging
import os

import requests
from django.conf import settings

duckling_url = settings.NLP_CONFIG.get('DUCKLING_URL')
language = settings.NLP_CONFIG.get('DUCKLING_LANGUAGE', 'en_US')

def get(text):
    """
    Makes a duckling request for text entities.
    :param text: Text to be parsed by Duckling.
    :return: Json returned by Duckling. Empty on error.
    """
    if duckling_url:
        logging.info('Duckling request')
        payload = {
            'locale': language,
            'text': text
        }
        try:
            resp = requests.post(duckling_url + "/parse", data=payload)
            if resp.status_code == 200:
                js = resp.json()
                print('Duckling:', js)
                if js is not None:
                    return to_entities(js)
            else:
                resp.raise_for_status()
        except Exception as ex:
            logging.exception('Exception @ Duckling')
    return {}


def to_entities(js):
    """Converts duckling output to the correct format."""
    entities = {}
    for entity in js:
        key, value = entity['dim'], entity['value']
        if key == 'time': key = 'datetime'
        if key not in entities:
            entities[key] = []
        entities[key].append(value)
    return entities



import os

from golem.nlu.utils import data_dir


class FuzzyMatcher:
    def __init__(self, entity):
        self.entity = entity
        pass

    def train(self, training_data):
        for obj in training_data['data']:
            label = obj['label']
            examples = obj['examples']
            # TODO generate all words in edit distance and save them



import os
import unidecode
import yaml

from golem.nlu import utils
from golem.nlu.cleanup import tokenize
# TODO remove character accents
from golem.nlu.nn.abs_model import NLUModel


class TrieKeywordModel(NLUModel):

    def __init__(self, entity, base_dir, is_training=False):
        self.entity = entity
        self.base_dir = base_dir

        if not is_training:
            with open(os.path.join(base_dir, "trie.json"), 'r') as f:
                self.trie = yaml.load(f)
            with open(os.path.join(base_dir, "metadata.json"), 'r') as f:
                metadata = yaml.load(f)
                self.should_stem = metadata.get('stemming', False)
                self.language = metadata.get('language', utils.get_default_language())

    def predict(self, utterance: str, threshold=None):
        return self.keyword_search(utterance)

    def prepare(self, examples, metadata):

        self.should_stem = metadata.get('stemming', False)
        self.language = metadata.get('language', utils.get_default_language())

        trie = self.prepare_keywords(examples)

        with open(os.path.join(self.base_dir, "trie.json"), "w") as f:
            yaml.dump(trie, f)

    def prepare_keywords(self, data) -> dict:
        """
        Prepares keywords for fast search.
        First words of each expression are processed into a trie.
        :param data:            A dict containing keyword data
        :param should_stem:     Boolean, whether to use stemmed keywords
                                for comparison. Default is false.
        :param language:        Stemming language. Default is 'en'.
        :return: trie ( dict )
        """
        # construct a trie from all chars of the keywords
        # may be thought of as finite state machine as well
        keywords = []
        root_state = {}
        for item in data:
            value = item.get('value')
            expressions = item.get('expressions', [])
            examples = item.get('examples', [])
            expressions += examples

            if not (value or expressions):
                raise ValueError("Keyword doesn't have any expressions")

            label = item.get('label', value)
            if not label:
                raise ValueError("Keyword doesn't have a label")

            if value:
                expressions.append(value)

            for expr in expressions:
                keywords.append((expr, label))

        for expression, label in keywords:
            expression = unidecode.unidecode(expression)
            words = tokenize(expression, self.should_stem, language=self.language)
            words = [str(word) for word in words]
            print(words)
            first_word = words[0]
            chars = list(first_word)
            state = root_state
            for idx, char in enumerate(chars):
                transitions = state.setdefault('transitions', {})
                state = transitions.setdefault(str(char), {})

            outputs = state.setdefault('outputs', [])
            outputs.append({
                "requires": words[1:],
                "label": label
            })

        return root_state

    def keyword_search(self, text) -> list:
        """
        Searches for keyword matches in text.
        :param text:        Text that should be searched
        :param should_stem: Boolean, whether to compare stemmed words.
                            Default is false.
        :param language:    Language for stemming. Default is 'en'.
        :return: A list of extracted entities, or an empty list.
        """
        extracted = set()
        text = unidecode.unidecode(text)
        words = tokenize(text, self.should_stem, self.language)  # TODO avoid repeating
        for idx, word in enumerate(words):
            state = self.trie
            chars = list(word)
            for char in chars:
                state = state.get('transitions', {}).get(char)
                if not state:
                    break  # unknown word
            if state:
                outputs = state.get('outputs', [])
                outputs.sort(key=lambda x: len(x.get('requires', [])), reverse=True)
                following = set(words[idx + 1:])
                for output in outputs:
                    required = set(output.get('requires', []))
                    if required & following == required:
                        extracted.add(output.get('label'))

        return [{"value": label} for label in extracted]



import json

from django.db import models


class WebMessageData(models.Model):
    uid = models.TextField()
    is_response = models.BooleanField()  # is this a response from the bot?
    timestamp = models.DateTimeField(auto_now=True)
    data = models.TextField()

    data_json = None

    def message(self):
        if not self.data_json:
            self.data_json = json.loads(self.data)
        return self.data_json



# Generated by Django 2.0.2 on 2018-06-05 10:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('golem', '0003_auto_20180207_2331'),
    ]

    operations = [
        migrations.CreateModel(
            name='WebMessageData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uid', models.TextField()),
                ('is_response', models.BooleanField()),
                ('timestamp', models.DateTimeField(auto_now=True)),
                ('data', models.TextField()),
            ],
        ),
    ]



# Copyright (c) 2008 Michael Dirolf (mike at dirolf dot com)

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

"""pyporter2: An implementation of the Porter2 stemming algorithm.

See http://snowball.tartarus.org/algorithms/english/stemmer.html"""
import re

regexp = re.compile(r"[^aeiouy]*[aeiouy]+[^aeiouy](\w*)")


def get_r1(word):
    # exceptional forms
    if word.startswith('gener') or word.startswith('arsen'):
        return 5
    if word.startswith('commun'):
        return 6

    # normal form
    match = regexp.match(word)
    if match:
        return match.start(1)
    return len(word)


def get_r2(word):
    match = regexp.match(word, get_r1(word))
    if match:
        return match.start(1)
    return len(word)


def ends_with_short_syllable(word):
    if len(word) == 2:
        if re.match(r"^[aeiouy][^aeiouy]$", word):
            return True
    if re.match(r".*[^aeiouy][aeiouy][^aeiouywxY]$", word):
        return True
    return False


def is_short_word(word):
    if ends_with_short_syllable(word):
        if get_r1(word) == len(word):
            return True
    return False


def remove_initial_apostrophe(word):
    if word.startswith("'"):
        return word[1:]
    return word


def capitalize_consonant_ys(word):
    if word.startswith('y'):
        word = 'Y' + word[1:]
    return re.sub(r"([aeiouy])y", '\g<1>Y', word)


def step_0(word):
    if word.endswith("'s'"):
        return word[:-3]
    if word.endswith("'s"):
        return word[:-2]
    if word.endswith("'"):
        return word[:-1]
    return word


def step_1a(word):
    if word.endswith('sses'):
        return word[:-4] + 'ss'
    if word.endswith('ied') or word.endswith('ies'):
        if len(word) > 4:
            return word[:-3] + 'i'
        else:
            return word[:-3] + 'ie'
    if word.endswith('us') or word.endswith('ss'):
        return word
    if word.endswith('s'):
        preceding = word[:-1]
        if re.search(r"[aeiouy].", preceding):
            return preceding
        return word
    return word


def step_1b(word, r1):
    if word.endswith('eedly'):
        if len(word) - 5 >= r1:
            return word[:-3]
        return word
    if word.endswith('eed'):
        if len(word) - 3 >= r1:
            return word[:-1]
        return word

    def ends_with_double(word):
        doubles = ['bb', 'dd', 'ff', 'gg', 'mm', 'nn', 'pp', 'rr', 'tt']
        for double in doubles:
            if word.endswith(double):
                return True
        return False

    def step_1b_helper(word):
        if word.endswith('at') or word.endswith('bl') or word.endswith('iz'):
            return word + 'e'
        if ends_with_double(word):
            return word[:-1]
        if is_short_word(word):
            return word + 'e'
        return word

    suffixes = ['ed', 'edly', 'ing', 'ingly']
    for suffix in suffixes:
        if word.endswith(suffix):
            preceding = word[:-len(suffix)]
            if re.search(r"[aeiouy]", preceding):
                return step_1b_helper(preceding)
            return word

    return word


def step_1c(word):
    if word.endswith('y') or word.endswith('Y'):
        if word[-2] not in 'aeiouy':
            if len(word) > 2:
                return word[:-1] + 'i'
    return word


def step_2(word, r1):
    def step_2_helper(end, repl, prev):
        if word.endswith(end):
            if len(word) - len(end) >= r1:
                if prev == []:
                    return word[:-len(end)] + repl
                for p in prev:
                    if word[:-len(end)].endswith(p):
                        return word[:-len(end)] + repl
            return word
        return None

    triples = [('ization', 'ize', []),
               ('ational', 'ate', []),
               ('fulness', 'ful', []),
               ('ousness', 'ous', []),
               ('iveness', 'ive', []),
               ('tional', 'tion', []),
               ('biliti', 'ble', []),
               ('lessli', 'less', []),
               ('entli', 'ent', []),
               ('ation', 'ate', []),
               ('alism', 'al', []),
               ('aliti', 'al', []),
               ('ousli', 'ous', []),
               ('iviti', 'ive', []),
               ('fulli', 'ful', []),
               ('enci', 'ence', []),
               ('anci', 'ance', []),
               ('abli', 'able', []),
               ('izer', 'ize', []),
               ('ator', 'ate', []),
               ('alli', 'al', []),
               ('bli', 'ble', []),
               ('ogi', 'og', ['l']),
               ('li', '', ['c', 'd', 'e', 'g', 'h', 'k', 'm', 'n', 'r', 't'])]

    for trip in triples:
        attempt = step_2_helper(trip[0], trip[1], trip[2])
        if attempt:
            return attempt

    return word


def step_3(word, r1, r2):
    def step_3_helper(end, repl, r2_necessary):
        if word.endswith(end):
            if len(word) - len(end) >= r1:
                if not r2_necessary:
                    return word[:-len(end)] + repl
                else:
                    if len(word) - len(end) >= r2:
                        return word[:-len(end)] + repl
            return word
        return None

    triples = [('ational', 'ate', False),
               ('tional', 'tion', False),
               ('alize', 'al', False),
               ('icate', 'ic', False),
               ('iciti', 'ic', False),
               ('ative', '', True),
               ('ical', 'ic', False),
               ('ness', '', False),
               ('ful', '', False)]

    for trip in triples:
        attempt = step_3_helper(trip[0], trip[1], trip[2])
        if attempt:
            return attempt

    return word


def step_4(word, r2):
    delete_list = ['al', 'ance', 'ence', 'er', 'ic', 'able', 'ible', 'ant', 'ement', 'ment', 'ent', 'ism', 'ate', 'iti',
                   'ous', 'ive', 'ize']

    for end in delete_list:
        if word.endswith(end):
            if len(word) - len(end) >= r2:
                return word[:-len(end)]
            return word

    if word.endswith('sion') or word.endswith('tion'):
        if len(word) - 3 >= r2:
            return word[:-3]

    return word


def step_5(word, r1, r2):
    if word.endswith('l'):
        if len(word) - 1 >= r2 and word[-2] == 'l':
            return word[:-1]
        return word

    if word.endswith('e'):
        if len(word) - 1 >= r2:
            return word[:-1]
        if len(word) - 1 >= r1 and not ends_with_short_syllable(word[:-1]):
            return word[:-1]

    return word


def normalize_ys(word):
    return word.replace('Y', 'y')


exceptional_forms = {'skis': 'ski',
                     'skies': 'sky',
                     'dying': 'die',
                     'lying': 'lie',
                     'tying': 'tie',
                     'idly': 'idl',
                     'gently': 'gentl',
                     'ugly': 'ugli',
                     'early': 'earli',
                     'only': 'onli',
                     'singly': 'singl',
                     'sky': 'sky',
                     'news': 'news',
                     'howe': 'howe',
                     'atlas': 'atlas',
                     'cosmos': 'cosmos',
                     'bias': 'bias',
                     'andes': 'andes'}

exceptional_early_exit_post_1a = ['inning', 'outing', 'canning', 'herring', 'earring', 'proceed', 'exceed', 'succeed']


def stem(word):
    """The main entry point in the old version of the API."""
    raise DeprecationWarning('stem() is deprecated starting with v1.0.0')


def algorithms():
    """Get a list of the names of the available stemming algorithms.

    The only algorithm currently supported is the "english", or porter2,
    algorithm.
    """
    return ['english']


def version():
    """Get the version number of the stemming module.

    This is the version number of the Stemmer module as a whole (not for an
    individual algorithm).
    """
    return '1.0.0'


class Stemmer:
    """An instance of a stemming algorithm.

    When creating a Stemmer object, there is one required argument:
    the name of the algorithm to use in the new stemmer. A list of the
    valid algorithm names may be obtained by calling the algorithms()
    function in this module. In addition, the appropriate stemming algorithm
    for a given language may be obtained by using the 2 or 3 letter ISO 639
    language codes.
    """
    max_cache_size = 10000

    def __init__(self, algorithm, cache_size=None):
        if algorithm not in ['english', 'eng', 'en']:
            raise KeyError("Stemming algorithm '%s' not found" % algorithm)
        if cache_size:
            self.max_cache_size = cache_size

    def stemWord(self, word):
        """Stem a word.

        This takes a single argument, word, which should either be a UTF-8
        encoded string, or a unicode object.

        The result is the stemmed form of the word. If the word supplied
        was a unicode object, the result will be a unicode object: if the
        word supplied was a string, the result will be a UTF-8 encoded string.
        """
        return Stemmer._stem(word)

    def stemWords(self, words):
        """Stem a list of words.

        This takes a single argument, words, which must be a sequence,
        iterator, generator or similar.

        The entries in words should either be UTF-8 encoded strings,
        or a unicode objects.

        The result is a list of the stemmed forms of the words. If the word
        supplied was a unicode object, the stemmed form will be a unicode
        object: if the word supplied was a string, the stemmed form will
        be a UTF-8 encoded string.
        """
        return [self.stemWord(word) for word in words]

    @classmethod
    def _stem(cls, word):

        if len(word) <= 2:
            return word
        word = remove_initial_apostrophe(word)

        # handle some exceptional forms
        if word in exceptional_forms:
            return exceptional_forms[word]

        word = capitalize_consonant_ys(word)
        r1 = get_r1(word)
        r2 = get_r2(word)
        word = step_0(word)
        word = step_1a(word)

        # handle some more exceptional forms
        if word in exceptional_early_exit_post_1a:
            return word

        word = step_1b(word, r1)
        word = step_1c(word)
        word = step_2(word, r1)
        word = step_3(word, r1, r2)
        word = step_4(word, r2)
        word = step_5(word, r1, r2)
        word = normalize_ys(word)

        return word



import logging
import time

from django.conf import settings

from golem.core.logging.abs_logger import MessageLogger


# ! DON'T IMPORT THIS FILE FROM settings.py !


class MessageLogging:

    def __init__(self, dialog):
        self.loggers = []
        self.dialog = dialog

    def log_user_message(self, type_, entities, accepted_time, accepted_state):
        """
        Log user message to all registered loggers.
        :param type_:           One of: message, postback, schedule
        :param entities:        A dict containing all the message entities
        :param accepted_time:   Unix timestamp - when was the message processed
        :param accepted_state:  Conversation state after processing the message
        :return:
        """
        # TODO this has to be called async
        message_text = entities.get("_message_text", "<POSTBACK/SCHEDULE>")
        for logger in MESSAGE_LOGGERS:
            logger.log_user_message(self.dialog, accepted_time, accepted_state, message_text, type_, entities)

    def log_bot_message(self, message, state):
        for logger in MESSAGE_LOGGERS:
            logger.log_bot_message(self.dialog, int(time.time()), state, message)

    def log_error(self, exception, state):
        for logger in MESSAGE_LOGGERS:
            logger.log_error(self.dialog, state, exception)

    def log_user(self, chat_session):
        for logger in MESSAGE_LOGGERS:
            logger.log_user(self.dialog, chat_session)


MESSAGE_LOGGERS = []


def register_logger(logger):
    if isinstance(logger, MessageLogger):
        logging.debug("Registering logger %s", logger.__class__.__name__)
        MESSAGE_LOGGERS.append(logger)
    else:
        raise ValueError("Error: Must be an instance of golem.core.abs_logger.MessageLogger")


try:
    for item in settings.GOLEM_CONFIG.get("MESSAGE_LOGGERS", []):
        register_logger(item)
except Exception as ex:
    raise ValueError("Error registering message loggers, is your configuration correct?") from ex



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2018-01-09 13:24
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('golem', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Chat',
            fields=[
                ('chat_id', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('user_uid', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='golem.User')),
            ],
        ),
    ]



from django import forms


class MessageForm(forms.Form):
    message = forms.CharField(
        label='Message',
        max_length=1024,
        widget=forms.TextInput(attrs={'placeholder': 'Type message here'})
    )



from golem.core.parsing.entity_extractor import EntityExtractor


class GolemExtractor(EntityExtractor):

    def __init__(self):
        super().__init__()
        self.nlu = None

    def extract_entities(self, text: str, max_retries=1):
        if not self.nlu:
            from golem.nlu.predict import GolemNLU
            self.nlu = GolemNLU()
            global GOLEM_NLU
            GOLEM_NLU = self.nlu

        # TODO use a separate thread (pool) to remove TF memory overhead
        return self.nlu.parse(text)


GOLEM_NLU = None



from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Parses a sentence using NLP module'

    def add_arguments(self, parser):
        parser.add_argument('text', nargs=1, type=str)

    def handle(self, *args, **options):
        from golem.nlp.predict import GolemNLU
        from pprint import pprint
        text = options['text'][0]
        if isinstance(text, str):
            nlu = GolemNLU()
            entities = nlu.parse(text)
            pprint(entities)
        else:
            raise ValueError('Text must be a string')



import pickle

import os
from keras.callbacks import EarlyStopping
from keras.layers import *
from keras.models import Sequential, load_model
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.preprocessing import LabelBinarizer
from sklearn.utils import shuffle

from golem.nlu import cleanup
from golem.nlu.nn.abs_model import NLUModel


class ContextualModel(NLUModel):
    def __init__(self, entity, base_dir, is_training=False):
        if not base_dir or not entity:
            raise ValueError('Base dir and entity name must be set')

        self.base_dir = base_dir
        self.entity = entity

        if not is_training:
            self._load_model()

    def _save_model(self):
        print("Saving model ...")
        with open(os.path.join(self.base_dir, "vocab.pkl"), "wb") as f:
            pickle.dump((self.vocab, self.labels, self.word_enc, self.label_enc), f)
        self.model.save(os.path.join(self.base_dir, "model.h5"))

    def _load_model(self):
        with open(os.path.join(self.base_dir, "vocab.pkl"), "rb") as f:
            self.vocab, self.labels, self.word_enc, self.label_enc = pickle.load(f)
        self.model = load_model(os.path.join(self.base_dir, "model.h5"))

    def _get_features(self, token, abs_idx, tokens):
        sentence_pos = 0  # abs_idx / len(tokens)  # [0;1]

        lwindow = tokens[max(0, abs_idx - 3): abs_idx]  # yes I know, 3 left, 2 right
        rwindow = tokens[abs_idx + 1: min(len(tokens), abs_idx + 3)]
        lcontext = np.sum(self.word_enc.transform(lwindow), axis=0)
        rcontext = np.sum(self.word_enc.transform(rwindow), axis=0)

        lcontext = np.asarray(lcontext)[0]
        rcontext = np.asarray(rcontext)[0]

        features = np.concatenate([[sentence_pos], lcontext, rcontext])
        #     print("Context:", lcontext, rcontext)

        return features

    def _make_xy(self, training_data):
        x, y = [], []
        for example in training_data:
            tokens = cleanup.tokenize(example[0])
            # take each entity for prediction
            # everything else will be negative samples
            idx = 0
            for abs_idx, token in enumerate(tokens):
                if token == '%':
                    idx += 1
                    label = example[idx]
                else:
                    label = 'none'

                features = self._get_features(token, abs_idx, tokens)
                label_encoded = self.label_enc.transform([label])

                x.append(features)
                y.append(label_encoded[0])

        return np.array(x), np.array(y)

    def train(self, training_data):
        # with open("data/nlp/training_data/context_test.yaml") as f:
        #     training_file = yaml.load(f)
        # training_data = ['data']

        self.vocab = set()
        self.labels = set()

        for example in training_data:
            tokens = cleanup.tokenize(example[0])
            self.vocab = self.vocab.union(tokens)
            self.labels = self.labels.union(example[1:])

        self.labels.add("none")

        self.word_enc = CountVectorizer()
        self.word_enc.fit(self.vocab)

        self.label_enc = LabelBinarizer()
        self.label_enc.fit(list(self.labels))

        print("Vocab:", self.vocab)
        print("Labels:", self.labels)

        x, y = self._make_xy(training_data)

        self.model = Sequential()
        self.model.add(Dense(256, input_shape=[x.shape[1]]))
        self.model.add(Activation('relu'))
        self.model.add(Dense(256))
        self.model.add(Activation('relu'))
        self.model.add(Dense(3))
        self.model.add(Activation('softmax'))

        x, y = shuffle(x, y)

        # TODO (1) try embeddings
        # TODO (2) try RNN, not MLP

        self.model.compile(
            optimizer='adam',
            loss='categorical_crossentropy',
            metrics=['accuracy']
        )
        self.model.fit(x, y, epochs=1000, callbacks=[EarlyStopping('loss', 0.001)])

        self._save_model()

    def test(self):
        pass  # TODO

    def predict(self, utterance, threshold=None):
        tokens = cleanup.tokenize(utterance)
        results = []
        current_entity = []
        current_label = None
        for idx, token in enumerate(tokens):
            features = self._get_features(token, idx, tokens)
            pred = self.model.predict(np.array([features]))
            label = self.label_enc.inverse_transform(pred)[0]
            if label == 'none':
                label = None

            if label:
                print("{} is {}".format(token, label))

            # TODO is this correct?

            if label and (not current_label or label == current_label):
                current_entity.append((label, idx, token))
                current_label = label
            elif current_entity:
                entity_text = ' '.join([x[2] for x in current_entity])
                start_idx = current_entity[0][1]
                end_idx = current_entity[-1][1]
                role = current_entity[0][0]
                result = {
                    "value": entity_text, "role": role,
                    "start_pos": start_idx, "end_pos": end_idx,
                }
                results.append(result)
                current_entity = []
                current_label = None
                if label:
                    current_entity.append((label, idx, token))
                    current_label = label

        if current_entity:
            entity_text = ' '.join([x[2] for x in current_entity])
            start_idx = current_entity[0][1]
            end_idx = current_entity[-1][1]
            role = current_entity[0][0]
            result = {
                "value": entity_text, "role": role,
                "start_pos": start_idx, "end_pos": end_idx,
            }
            results.append(result)

        return results



from celery import shared_task


@shared_task
def on_message(uid, chat_id, message_text, dialog, from_user):
    from golem.models import User, Chat, Message

    if chat_id is None:
        chat_id = uid
    user, was_created = User.objects.get_or_create(uid=uid, defaults={"uid": uid})
    if was_created:
        # TODO save user profile etc.
        user.save()
    chat, was_created = Chat.objects.get_or_create(chat_id=chat_id, defaults={"chat_id": chat_id, "user_uid": user})
    if was_created:
        chat.save()
    message = Message()
    message.chat = chat
    message.text = message_text
    message.is_from_user = from_user
    message.intent = dialog.context.get("intent", max_age=0)
    message.state = dialog.current_state_name
    message.save()



import redis
from django.conf import settings

_connection_pool = None
_redis = None


def get_redis():
    global _connection_pool
    global _redis
    if not _connection_pool:
        config = settings.GOLEM_CONFIG.get('REDIS')
        _connection_pool = redis.ConnectionPool(
            host=config['HOST'],
            port=config['PORT'],
            password=config['PASSWORD'],
            db=0,
            max_connections=2
        )
    if not _redis:
        _redis = redis.StrictRedis(connection_pool=_connection_pool)
    return _redis



# Based on: http://norvig.com/spell-correct.html


def nearest_word(word, glove):
    corrections = known(word, glove)
    if corrections:
        return min(corrections, key=lambda x: distance(x, word))
    return None


def known(word, glove):
    edits_1 = edits(word)
    edits_2 = []  # (e2 for e1 in edits_1 for e2 in edits(e1))
    return set(w for w in edits_1.union(edits_2) if glove.contains(w))


def distance(m, n):  # TODO something more clever
    score = 0
    for i in range(min(len(m), len(n))):
        score += abs(ord(m[i]) - ord(n[i]))
    score += (len(m) - len(n)) ** 2
    return score


def edits(word):  # TODO something better
    letters = '1234567890aábcčdďeéěfghchiíjklmnňoópqrřsštťuúůvwxyýzž'
    splits = [(word[:i], word[i:]) for i in range(len(word) + 1)]
    deletes = [L + R[1:] for L, R in splits if R]
    transposes = [L + R[1] + R[0] + R[2:] for L, R in splits if len(R) > 1]
    replaces = [L + c + R[1:] for L, R in splits if R for c in letters]
    inserts = [L + c + R for L, R in splits for c in letters]
    edited_1 = set(deletes + transposes + replaces + inserts)
    return edited_1


similar_letters = {
    "a": "ásdzxqw",
    "b": "vnfghj",
    "c": "čxdfv",
    "d": "ďswerfxc",
    "e": "éwsdfr",
    "f": "rdcvgt",
    "g": "tfvbht",
    "h": "ygbnju",
    "i": "íujklo",
    "j": "uhnmki",
    "k": "ijmlo",
    "l": "ĺľpokmů",
    "m": "njkl",
    "n": "ňmjhb",
    "o": "óôiklp",
    "p": "olúô",
    "q": "asw",
    "r": "ŕředfgt",
    "s": "šazxdewq",
    "t": "ťrfgy",
    "u": "úůhjki",
    "v": "cfgb",
    "w": "qasde",
    "x": "zsdc",
    "y": "ýtghuj",
    "z": "žasxcd"
}



# Generated by Django 2.0.2 on 2018-02-07 22:31

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('golem', '0002_chat'),
    ]

    operations = [
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('text', models.TextField(blank=True, null=True)),
                ('is_from_user', models.BooleanField()),
                ('time', models.DateTimeField(auto_now=True)),
                ('intent', models.TextField(blank=True, db_index=True, max_length=255, null=True)),
                ('state', models.TextField(blank=True, db_index=True, max_length=255, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='QuickReply',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('text', models.TextField(max_length=255)),
                ('message', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='golem.Message')),
            ],
        ),
        migrations.RemoveField(
            model_name='user',
            name='last_active',
        ),
        migrations.AddField(
            model_name='message',
            name='chat',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='golem.Chat'),
        ),
    ]



import json

import numpy as np
import os


class Embedding:

    def __init__(self, filename, max_cache_size=1024):

        # TODO use R-tree for similarity queries

        if not os.path.exists(filename):
            raise FileNotFoundError("The embedding file {} doesn't exist.".format(filename))

        self.fp = open(filename, 'rb')
        self.dimension = int(self.fp.readline().split()[1])
        self.max_cache_size = max_cache_size

        # cache word offsets in the embedding file for quick lookup
        pos_path = filename + ".offsets"
        if os.path.exists(pos_path):
            with open(pos_path, 'rb') as pos_f:
                self.positions = json.load(pos_f)
        else:
            with open(pos_path, 'wb') as pos_f:
                # FIXME open exclusively, other celery thread might screw up
                self.positions = self.cache_positions()
                json.dump(pos_f, self.positions)

        # LRU cache to limit disk reads
        self.cache = {}
        self.recent = []

    def cache_positions(self):

        self.fp.seek(0)
        positions = {}

        line, off = self.fp.readline(), self.fp.tell()
        while line:
            word = line.split(maxsplit=1)[0]
            word = word.lower()
            # TODO there will be lower-upper duplicates, but better than missing words
            positions[word] = off
            off = self.fp.tell()
            line = self.fp.readline()

        return positions

    def word2vec(self, word):

        if not isinstance(word, str):
            raise ValueError("Word must be a string, but is {}".format(type(word)))

        word = word.lower()

        # search in cache
        if word in self.cache:
            self.recent.remove(word)
            self.recent.insert(0, word)
            return self.cache[word]

        # search in embedding file
        if word in self.positions:
            off = self.positions[word]
            self.fp.seek(off)
            line = self.fp.readline().split()
            vector = np.array([float(x) for x in line[1:]])

            if word == line[0]:

                # remove oldest cached item if cache is full
                if len(self.recent) >= self.max_cache_size:
                    removed_word = self.recent.pop()
                    del self.cache[removed_word]

                # add this word to cache
                self.recent.insert(0, word)
                self.cache[word] = vector

                return vector
            else:
                raise Exception("Word2vec cache is corrupt, have you modified the embedding file?")

        # TODO construct from sub-word vectors

        return None



import json

from golem.core.responses.responses import MessageElement, _get_payload_string
from golem.core.serialize import json_serialize


class QuickReply(MessageElement):
    """
    A button with text that gets sent as message when clicked.
    """

    def __init__(self, title=None, payload=None, image_url=None):
        self.title = title
        self.payload = payload
        self.image_url = image_url

    def to_response(self):
        response = {
            "content_type": 'text',
            'title': self.title[:20],
            'payload': json.dumps(self.payload if self.payload else {}, default=json_serialize)
        }
        if self.image_url:
            response['image_url'] = self.image_url
        return response

    def __str__(self):
        text = 'quick_reply: '
        if self.title:
            text += self.title + ': '
        if self.payload:
            text += _get_payload_string(self.payload)
        return text


class LocationQuickReply(QuickReply):
    """
    A button that asks for user's location when clicked.
    Currently only supported on Facebook Messenger.
    """

    def __init__(self):
        super().__init__()

    def to_response(self):
        return {'content_type': 'location'}

    def __str__(self):
        return str(super) + " location"



import tensorflow as tf



class CharEntityExtractor:

    def __init__(self, entity_name, model_dir):
        self.entity_name = entity_name
        self.model_dir = model_dir
        self.max_chars = 320
        self.batch_size = 32
        self.initial_step = 1e-3
        self.num_units = 32
        self.session = tf.Session()

    def create_network(self):
        with tf.name_scope(self.entity_name):
            self.x = tf.placeholder(tf.int8, [self.batch_size, self.max_chars, 1], name="x")
            self.y = tf.placeholder(tf.float16, [self.batch_size, self.max_chars, 1], name="y")

            with tf.name_scope("encoder_fwd"):
                self.cell_fwd = tf.nn.rnn_cell.GRUCell(
                    num_units=self.num_units,
                    activation=tf.nn.relu,
                )

            with tf.name_scope("encoder_bwd"):
                self.cell_bwd = tf.nn.rnn_cell.GRUCell(
                    num_units=self.num_units,
                    activation=tf.nn.relu,
                )

            with tf.name_scope("encoder"):
                self.encoder_out, self.encoder_state = tf.nn.bidirectional_dynamic_rnn(
                    cell_fw=self.cell_fwd,
                    cell_bw=self.cell_bwd,
                    inputs=self.x,
                    dtype=tf.float16,
                )

            self.y_ = None



import json

import os
from django.db import models

strategies = ["trait", "keywords"]
languages = ["en", "cz"]


class Entity:  # (models.Model):
    name = models.CharField(max_length=100)
    strategy = models.CharField(max_length=100, choices=strategies)
    stemming = models.BooleanField(default=False)
    language = models.CharField(max_length=10, choices=languages, default="en")
    threshold = models.FloatField(default=0.7)

    @staticmethod
    def fromFile(path):
        with open(path) as f:
            data = json.load(f)  # type: dict
        filename = os.path.split(path)[-1]
        name, ext = os.path.splitext(filename)
        e = Entity()
        e.name = name
        e.strategy = data["strategy"]
        e.stemming = data.get("stemming", False)
        e.language = data.get("language", "en")
        e.threshold = data.get("threshold", 0.7)
        values = data.get("data", [])
        e.values = []
        for value in values:
            v = EntityValue()
            v.name = value.get('label', value.get('value'))
            e.values.append(v)
        return e

    def get_values(self):
        if hasattr(self, 'values'):
            return self.values
        return []

    def __str__(self):
        return self.name


class EntityValue:  # (models.Model):
    name = models.CharField(max_length=100)



from abc import ABC, abstractmethod


class EntityExtractor(ABC):
    """
    Abstract class for entity extractors.
    Responsible for processing text and extracting entities such as names, dates, places etc.
    """

    def __init__(self):
        pass

    @abstractmethod
    def extract_entities(self, text: str, max_retries=5):
        """
        Extracts entities from text (a message from user).
        :param text:        a string.
        :param max_retries: how many times to retry on error.
        :return: A dict of extracted entities.
                 For example:
                 {
                    "place": [
                        {"value": "Prague"},
                        {"value": "New York", "metadata": {}}
                    ],
                    "name": [{"value": "Golem"}],
                    ...
                 }
        """
        return dict()



import json
import logging
import time

import importlib
import re
from django.conf import settings

from golem.core.chat_session import ChatSession
from golem.core.responses import LinkButton
from golem.core.responses.responses import TextMessage
from golem.tasks import accept_inactivity_callback, accept_schedule_callback
from .context import Context
from .flow import Flow
from .logger import MessageLogging
from .persistence import get_redis
from .serialize import json_deserialize, json_serialize
from .tests import ConversationTestRecorder


class DialogManager:
    version = '1.30'

    def __init__(self, session: ChatSession):
        self.session = session
        self.uid = session.chat_id  # for backwards compatibility
        # self.logger = ElasticsearchLogger(session)
        self.logger = MessageLogging(self)
        self.db = get_redis()
        self.log = logging.getLogger()

        self.should_log_messages = settings.GOLEM_CONFIG.get('SHOULD_LOG_MESSAGES', False)
        self.error_message_text = settings.GOLEM_CONFIG.get('ERROR_MESSAGE_TEXT')

        entities = {}
        version = self.db.get('dialog_version')
        self.log.info('Initializing dialog for chat %s...' % session.chat_id)
        self.current_state_name = None
        self.context = None  # type: Context
        if version and \
                version.decode('utf-8') == DialogManager.version and \
                self.db.hexists('session_entities', self.session.chat_id):
            entities_string = self.db.hget('session_entities', self.session.chat_id)
            history_string = self.db.hget('session_history', self.session.chat_id)
            self.init_flows()
            state = self.db.hget('session_state', self.session.chat_id).decode('utf-8')
            counter = int(self.db.hget('session_counter', self.session.chat_id))
            self.log.info('Session exists at state %s' % (state))
            self.move_to(state, initializing=True)
            # self.log.info(entities_string)
            entities = json.loads(entities_string.decode('utf-8'), object_hook=json_deserialize)
            history = json.loads(history_string.decode('utf-8'), object_hook=json_deserialize)
        else:
            self.current_state_name = 'default.root'
            self.log.info('Creating new session...')
            counter = 0
            history = []
            self.init_flows()
            self.logger.log_user(self.session)
        self.context = Context(entities=entities, history=history, counter=counter, dialog=self)  # type: Context

    def init_flows(self):
        self.flows = {}
        flow_definitions = self.create_flows()
        for flow_name, flow_definition in flow_definitions.items():
            self.flows[flow_name] = Flow(flow_name, dialog=self, definition=flow_definition)
        self.current_state_name = 'default.root'

    def create_flows(self):
        flow = {}
        BOTS = settings.GOLEM_CONFIG.get('BOTS')
        for module_name in BOTS:
            module = importlib.import_module(module_name)
            flow.update(module.flow)
        return flow

    @staticmethod
    def clear_chat(chat_id):
        db = get_redis()
        db.hdel('session_state', chat_id)
        db.hdel('session_entities', chat_id)

    def process(self, message_type, entities):
        self.session.interface.processing_start(self.session)
        accepted_time = time.time()
        accepted_state = self.current_state_name
        # Only process messages and postbacks (not 'seen_by's, etc)
        if message_type not in ['message', 'postback', 'schedule']:
            return

        self.log.info('-- USER message ----------------------------------')

        # if message_type != 'schedule':
        self.context.counter += 1

        entities = self.context.add_entities(entities)

        if self.test_record_message(message_type, entities):
            return

        if message_type != 'schedule':
            self.save_inactivity_callback()

        self.log.info('++ PROCESSING ++++++++++++++++++++++++++++++++++++')

        if not self.check_state_transition():
            if not self.check_intent_transition():
                # run the action
                self.run_accept(save_identical=True)
                self.save_state()

        self.session.interface.processing_end(self.session)

        # leave logging message to the end so that the user does not wait
        self.logger.log_user_message(message_type, entities, accepted_time, accepted_state)

    def schedule(self, callback_name, at=None, seconds=None):
        self.log.info('Scheduling callback "{}": at {} / seconds: {}'.format(callback_name, at, seconds))
        if at:
            if at.tzinfo is None or at.tzinfo.utcoffset(at) is None:
                raise Exception('Use datetime with timezone, e.g. "from django.utils import timezone"')
            accept_schedule_callback.apply_async((self.session.to_json(), callback_name), eta=at)
        elif seconds:
            accept_schedule_callback.apply_async((self.session.to_json(), callback_name), countdown=seconds)
        else:
            raise Exception('Specify either "at" or "seconds" parameter')

    def inactive(self, callback_name, seconds):
        self.log.info('Setting inactivity callback "{}" after {} seconds'.format(callback_name, seconds))
        accept_inactivity_callback.apply_async(
            (self.session.to_json(), self.context.counter, callback_name, seconds),
            countdown=seconds)

    def save_inactivity_callback(self):
        self.db.hset('session_active', self.session.chat_id, time.time())
        callbacks = settings.GOLEM_CONFIG.get('INACTIVE_CALLBACKS')
        if not callbacks:
            return
        for name in callbacks:
            seconds = callbacks[name]
            self.inactive(name, seconds)

    def test_record_message(self, message_type, entities):
        record, record_age = self.context.get_age('test_record')
        self.recording = False
        if not record:
            return False
        if record_age == 0:
            if record == 'start':
                self.send_response(ConversationTestRecorder.record_start())
            elif record == 'stop':
                self.send_response(ConversationTestRecorder.record_stop())
            else:
                self.send_response("Use /test_record/start/ or /test_record/stop/")
            self.save_state()
            return True
        if record == 'start':
            ConversationTestRecorder.record_user_message(message_type, entities)
            self.recording = True
        return False

    def run_accept(self, save_identical=False):
        self.log.warn('Running ACCEPT action of {}'.format(self.current_state_name))
        state = self.get_state()
        if not state.accept:
            self.log.warn('State does not have an ACCEPT action, we are done.')
            return
        response, new_state_name = state.accept(
            state=state)  # FIXME <-- don't crash on invalid return value (not iterable)
        self.send_response(response)
        self.move_to(new_state_name, save_identical=save_identical)

    def run_init(self):
        self.log.warning('Running INIT action of {}'.format(self.current_state_name))
        state = self.get_state()
        if not state.init:
            self.log.warning('State does not have an INIT action, we are done.')
            return
        response, new_state_name = state.init(state=state)
        self.send_response(response)
        self.move_to(new_state_name)

    def check_state_transition(self):
        new_state_name = self.context.get('_state', max_age=0)
        return self.move_to(new_state_name)

    def check_intent_transition(self):
        intent = self.context.get('intent', max_age=0)
        if not intent:
            return False
        # Get custom intent transition
        new_state_name = self.get_state().get_intent_transition(intent)
        # If no custom intent transition present, move to the flow whose 'intent' field matches intent
        # Check accepted intent of the current flow's states
        if not new_state_name:
            for state in self.get_flow().states.values():
                if state.intent and re.match(state.intent, intent):
                    new_state_name = state.name
                    break
        # Check accepted intent of all flows
        if not new_state_name:
            for flow in self.flows.values():
                if re.match(flow.intent, intent):
                    new_state_name = flow.name + '.root'
                    break

        if not new_state_name:
            self.log.error('Error! Found intent "%s" but no flow present for it!' % intent)
            return False

        new_state_name = new_state_name + ':accept'
        self.log.info('Moving based on intent %s...' % (intent))
        return self.move_to(new_state_name)

    def get_flow(self, flow_name=None):
        if not flow_name:
            flow_name, _ = self.current_state_name.split('.', 1)
        return self.flows.get(flow_name)

    def get_state(self, flow_state_name=None):
        flow_name, state_name = (flow_state_name or self.current_state_name).split('.', 1)
        flow = self.get_flow(flow_name)
        return flow.get_state(state_name) if flow else None

    def move_to(self, new_state_name, initializing=False, save_identical=False):
        # if flow prefix is not present, add the current one
        action = 'init'
        if isinstance(new_state_name, int):
            new_state = self.context.get_history_state(new_state_name - 1)
            new_state_name = new_state['name'] if new_state else None
            action = None
        if not new_state_name:
            new_state_name = self.current_state_name
        if new_state_name.count(':'):
            new_state_name, action = new_state_name.split(':', 1)
        if ('.' not in new_state_name):
            new_state_name = self.current_state_name.split('.')[0] + '.' + new_state_name
        if not self.get_state(new_state_name):
            self.log.info('Error: State %s does not exist! Staying at %s.' % (new_state_name, self.current_state_name))
            return False
        identical = new_state_name == self.current_state_name
        if not initializing and (not identical or save_identical):
            self.context.add_state(new_state_name)
        if not new_state_name or identical:
            return False
        previous_state = self.current_state_name
        self.current_state_name = new_state_name
        if not initializing:
            self.log.info('MOVING %s -> %s %s' % (previous_state, new_state_name, action))

            # notify the interface that the state was changed
            self.session.interface.state_change(self.current_state_name)
            # record change if recording tests
            if self.recording:
                ConversationTestRecorder.record_state_change(self.current_state_name)

            try:

                if action == 'init':
                    self.run_init()
                elif action == 'accept':
                    self.run_accept()

            except Exception as e:
                logging.error('*****************************************************')
                logging.error('Exception occurred while running action {} of state {}'
                              .format(action, new_state_name))
                logging.error('Chat id: {}'.format(self.session.chat_id))
                try:
                    context_debug = self.get_state().dialog.context.debug()
                    logging.error('Context: {}'.format(context_debug))
                except:
                    pass
                logging.exception('Exception follows')
                if self.error_message_text:
                    self.send_response([self.error_message_text])

                # Raise the error if we are in a test
                if self.session.is_test:
                    raise e

        self.save_state()
        return True

    def save_state(self):
        if not self.context:
            return
        self.log.info('Saving state at %s' % (self.current_state_name))
        self.db.hset('session_state', self.session.chat_id, self.current_state_name)
        self.db.hset('session_history', self.session.chat_id, json.dumps(self.context.history, default=json_serialize))
        self.db.hset('session_entities', self.session.chat_id,
                     json.dumps(self.context.entities, default=json_serialize))
        self.db.hset('session_counter', self.session.chat_id, self.context.counter)
        self.db.hset('session_interface', self.session.chat_id, self.session.interface.name)
        self.db.set('dialog_version', DialogManager.version)

        # save chat session to redis, TODO
        session = json.dumps(self.session.to_json())
        self.db.hset("chat_session", self.session.chat_id, session)

    def send_response(self, responses):
        if not responses:
            return
        self.log.info('-- CHATBOT message -------------------------------')

        if not isinstance(responses, list):
            return self.send_response([responses])

        for response in responses:
            if isinstance(response, str):
                response = TextMessage(text=response)

            # Send the response
            self.session.interface.post_message(self.session, response)

            # Record if recording
            if self.recording:
                ConversationTestRecorder.record_bot_message(response)

        for response in responses:
            # Log the response
            #self.log.info('Message: {}'.format(response))
            self.logger.log_bot_message(response, self.current_state_name)

            # text = response.text if hasattr(response, 'text') else (response if isinstance(response, str) else None)
            # if text and self.should_log_messages:
            #     message_logger.on_message.delay(self.session, text, self, from_user=False)

    def dont_understand(self):
        # TODO log to chatbase
        # TODO work in progress
        from golem.core.parsing import golem_extractor
        utterance = self.context.get("_message_text", max_age=0)
        nlu = golem_extractor.GOLEM_NLU
        if not nlu or not utterance:
            print("NLU instance and message text can't be None")
            return
        intent = nlu.parse_entity(utterance, 'intent', threshold=0.5)
        if intent:
            text = "I'm not sure what you mean. Are you talking about \"{}\"?".format(intent[0]['value'])
            message = TextMessage(text).with_replies(['Yes', 'No'])
            self.send_response(message)
        else:
            text = "I'm not sure what you mean. Could you help me learn?"
            message = TextMessage(text).add_button(LinkButton("WebView", "http://zilinec.me/intent.html"))
            self.send_response(message)
            # TODO webview



from golem.core.chat_session import ChatSession, Profile
from golem.core.message_parser import parse_text_message
from golem.core.persistence import get_redis
from golem.core.responses import TextMessage
from golem.tasks import accept_user_message


class GoogleActionsInterface():
    name = 'google_actions'
    prefix = 'goog'
    messages = []
    states = []

    response_cache = {}

    @staticmethod
    def clear():
        GoogleActionsInterface.messages = []
        GoogleActionsInterface.states = []

    @staticmethod
    def fill_session_profile(session: ChatSession):
        if not session:
            raise ValueError("Session is None")
        session.profile = Profile(first_name='Test', last_name='')
        return session

    @staticmethod
    def post_message(session, response):
        GoogleActionsInterface.messages.append(response)
        get_redis().set("response_for_{id}".format(id=session.chat_id), str(response))  # TODO

    @staticmethod
    def convert_responses(session, responses):

        # TODO
        response = responses or "Hello world!"

        if isinstance(response, TextMessage):
            text = response.text
        elif isinstance(response, str):
            text = response
        else:
            text = "Dummy response"

        json_response = {
            "conversationToken": session.meta.get("chat_id"),
            "expectUserResponse": True,
            "expectedInputs": [
                {
                    "inputPrompt": {
                        "richInitialPrompt": {
                            "items": [
                                {
                                    "simpleResponse": {
                                        "textToSpeech": text,
                                        # "Howdy! I can tell you fun facts about almost any number, like 42. What do you have in mind?",
                                        "displayText": text
                                        # "Howdy! I can tell you fun facts about almost any number. What do you have in mind?"
                                    }
                                }
                            ],
                            "suggestions": []
                        }
                    },
                    "possibleIntents": [
                        {
                            "intent": "actions.intent.TEXT"
                        }
                    ]
                }
            ]
        }

        return json_response
        # return json.dumps(json_response)

    @staticmethod
    def send_settings(settings):
        pass

    @staticmethod
    def processing_start(session):
        pass

    @staticmethod
    def processing_end(session):
        pass

    @staticmethod
    def state_change(state):
        if not GoogleActionsInterface.states or GoogleActionsInterface.states[-1] != state:
            GoogleActionsInterface.states.append(state)

    @staticmethod
    def accept_request(body):
        uid = body['user']['userId']
        chat_id = body['conversation']['conversationId']
        meta = {"uid": uid, "chat_id": chat_id}
        profile = Profile(uid, None, None)
        session = ChatSession(GoogleActionsInterface, chat_id, meta, profile)
        accept_user_message.delay(session.to_json(), body).get()
        # responses = GoogleActionsInterface.response_cache.get(session.chat_id)
        responses = get_redis().get("response_for_{id}".format(id=session.chat_id))  # TODO
        return GoogleActionsInterface.convert_responses(session, responses.decode('utf8'))

    @staticmethod
    def parse_message(msg, num_tries=1):
        # intent = msg['inputs']['intent']
        # if intent == "assistant.intent.action.MAIN":
        #     intent = 'default'
        text = msg['inputs'][0]['rawInputs'][0]['query']
        parsed = parse_text_message(text)
        return parsed



import json
import logging

import requests

from golem.core.parsing import date_utils
from golem.core.parsing.entity_extractor import EntityExtractor


class DucklingExtractor(EntityExtractor):

    def __init__(self, url, lang='en_US'):
        super().__init__()
        if not url:
            raise ValueError("Duckling URL must be set")
        self.duckling_url = url
        self.language = lang

    def extract_entities(self, text: str, max_retries=1):
        """
        Makes a duckling request for text entities.
        :param text: Text to be parsed by Duckling.
        :return: Json returned by Duckling. Empty on error.
        """
        payload = {
            'locale': self.language,
            'text': text
        }
        try:
            resp = requests.post(self.duckling_url + "/parse", data=payload)
            if resp.status_code == 200:
                jsn = resp.json()
                logging.debug('Duckling:', jsn)
                if jsn is not None:
                    return self.to_entities(jsn)
            else:
                resp.raise_for_status()
        except Exception:
            if max_retries > 0:
                return self.extract_entities(text, max_retries - 1)
            else:
                logging.exception('Exception @ Duckling')
        return {}

    def to_entities(self, jsn):
        """Converts duckling output to the correct format."""
        entities = {}
        for entity in jsn:
            key, value = entity['dim'], entity['value']
            if key == 'time': key = 'datetime'
            if key not in entities:
                entities[key] = []
            entities[key].append(value)
        return self._process_wit_entities(entities)

    def _process_wit_entities(self, entities: dict):

        entities = self._process_metadata(entities)

        if 'datetime' in entities:
            datetime = entities['datetime']
            duration = entities.get('duration', None)
            append = date_utils.process_datetime(datetime, duration)
            entities.update(append)

        return entities

    def _process_metadata(self, entities: dict):
        for entity, values in entities.items():
            for value in values:
                # parse string metadata from Wit into a dict
                metadata = value.get('metadata')
                if metadata and isinstance(metadata, str):
                    try:
                        value['metadata'] = json.loads(metadata)
                    except:
                        self.log.warning("Ignoring invalid metadata for entity {}: {}".format(
                            entity, metadata
                        ))
                        value['metadata'] = None
        return entities
